"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


import numpy as np
import sympy as sp

from qrisp.arithmetic import (
    eq,
    geq,
    gt,
    inpl_mult,
    leq,
    lt,
    neq,
    polynomial_encoder,
    q_mult,
    sbp_add,
    quantum_bit_shift,
)
from qrisp.core import QuantumVariable
from qrisp.misc import gate_wrap


def signed_int_iso(x, n):
    if int(x) < -(2**n) or int(x) >= 2**n:
        raise Exception("Applying signed integer isomorphism resulted in overflow")

    if x >= 0:
        return x % 2**n
    else:
        return -abs(x) % 2 ** (n + 1)


def signed_int_iso_inv(y, n):
    y = y % 2 ** (n + 1)
    if y < 2**n:
        return y
    else:
        return -(2 ** (n + 1)) + y


# Truncates a polynomial of the form p(x) = 2**k_0*x*i_0 + 2**k_1*x**i_1 ...
# where every summand where the power of the coefficients does not lie in the interval
# trunc bounds is removed
def trunc_poly(poly, trunc_bounds):
    # Convert to sympy polynomial
    poly = sp.poly(poly)

    # Clip upper bound
    poly = poly.trunc(2.0 ** (trunc_bounds[1]))

    # Clip lower bound
    poly = poly / 2.0 ** trunc_bounds[0]
    poly = poly - sp.poly(poly).trunc(1)
    poly = poly * 2.0 ** trunc_bounds[0]

    return poly.expr.expand()


class QuantumFloat(QuantumVariable):
    r"""
    This subclass of :ref:`QuantumVariable` can represent floating point numbers
    (signed and unsigned) up to an arbitrary precision.

    The technical details of the employed arithmetic can be found in this
    `article <https://ieeexplore.ieee.org/document/9815035>`_.

    To create a QuantumFloat we call the constructor:

    >>> from qrisp import QuantumFloat
    >>> a = QuantumFloat(3, -1, signed = False)

    Here, the 3 indicates the amount of mantissa qubits and the -1 indicates the
    exponent.

    For unsigned QuantumFloats, the decoder function is given by

    .. math::

        f_{k}(i) = i2^{k}

    Where $k$ is the exponent.

    We can check which values can be represented:

    >>> for i in range(2**a.size): print(a.decoder(i))
    0.0
    0.5
    1.0
    1.5
    2.0
    2.5
    3.0
    3.5

    We see $2^3 = 8$ values, because we have 3 mantissa qubits. The exponent is -1,
    implying the precision is $0.5 = 2^{-1}$.

    For signed QuantumFloats, the decoder function is

    .. math::

        f_{k}^{n}(i) = \begin{cases} i2^{k} & \text{if } i < 2^n \\ (i - 2^{n+1})2^k &
        \text{else} \end{cases}

    Where $k$ is again, the exponent and $n$ is the mantissa size.


    Another example:

    >>> b = QuantumFloat(2, -2, signed = True)
    >>> for i in range(2**b.size): print(b.decoder(i))
    0.0
    0.25
    0.5
    0.75
    -1.0
    -0.75
    -0.5
    -0.25

    Here, we have $2^2 = 4$ values and their signed equivalents. Their precision is
    $0.25 = 2^{-2}$.


    **Arithmetic**

    Many operations known from classical arithmetic work for QuantumFloats in infix
    notation.

    Addition:

    >>> a[:] = 1.5
    >>> b[:] = 0.25
    >>> c = a + b
    >>> print(c)
    {1.75: 1.0}

    Subtraction:

    >>> d = a - c
    >>> print(d)
    {-0.25: 1.0}

    Multiplication:

    >>> e = d * b
    >>> print(e)
    {-0.0625: 1.0}

    And even division:

    >>> a = QuantumFloat(3)
    >>> b = QuantumFloat(3)
    >>> a[:] = 7
    >>> b[:] = 2
    >>> c = a/b
    >>> print(c)
    {3.5: 1.0}

    Floor division:

    >>> d = a//b
    >>> print(d)
    {3.0: 1.0}

    Inversion:

    >>> a = QuantumFloat(3, -1)
    >>> a[:] = 3.5
    >>> b= a**-1
    >>> print(b)
    {0.25: 1.0}

    Note that the latter is only an approximate result. This is because in many cases,
    the results of division can not be stored in a finite amount of qubits, forcing us
    to approximate.
    To get a better approximation we can use the :meth:`q_div <qrisp.q_div>` and
    :meth:`qf_inversion <qrisp.qf_inversion>` functions and specify the precision:

    >>> from qrisp import q_div, qf_inversion
    >>> a = QuantumFloat(3)
    >>> a[:] = 1
    >>> b = QuantumFloat(3)
    >>> b[:] = 7
    >>> c = q_div(a, b, prec = 6)
    >>> print(c)
    {0.140625: 1.0}

    Comparing with the classical result (0.1428571428):

    >>> 1/7 - 0.140625
    0.002232142857142849

    We see that the result is inside the expected precision of $2^{-6} =  0.015625$.


    **In-place Operations**

    Further supported operations are inplace addition, subtraction (with both classical
    and quantum values):

    >>> a = QuantumFloat(4, signed = True)
    >>> a[:] = 4
    >>> b = QuantumFloat(4)
    >>> b[:] = 3
    >>> a += b
    >>> print(a)
    {7: 1.0}
    >>> a -= 2
    >>> print(a)
    {5: 1.0}

    .. warning::
        Additions that would result in overflow, raise no errors. Instead, the additions
        are performed `modular <https://en.wikipedia.org/wiki/Modular_arithmetic>`_.

        >>> c = QuantumFloat(3)
        >>> c += 9
        >>> print(c)
        {1: 1.0}

    For inplace multiplications, only classical values are allowed:

    >>> a *= -3
    >>> print(a)
    {-15: 1.0}

    .. note::
        In-place multiplications can change the mantissa size to prevent overflow
        errors. If you want to prevent this behavior, look into
        :meth:`inpl_mult <qrisp.inpl_mult>`.

        >>> a.size
        7

    **Bitshifts**

    Bitshifts can be executed for free (i.e. not requiring any quantum gates). We can
    either use the :meth:`exp_shift <qrisp.QuantumFloat.exp_shift>` method or use the
    infix operators. Note that the bitshifts work in-place.


    >>> a.exp_shift(3)
    >>> print(a)
    {-120: 1.0}
    >>> a >> 5
    >>> print(a)
    {-3.75: 1.0}

    **Comparisons**

    QuantumFloats can be compared to Python floats using the established operators. The
    return values are :ref:`QuantumBools <QuantumBool>`:

    >>> from qrisp import h
    >>> a = QuantumFloat(4)
    >>> h(a[2])
    >>> print(a)
    {0: 0.5, 4: 0.5}
    >>> comparison_qbl_0 = (a < 4 )
    >>> print(comparison_qbl_0)
    {False: 0.5, True: 0.5}

    Comparison to other QuantumFloats also works:

    >>> b = QuantumFloat(3)
    >>> b[:] = 4
    >>> comparison_qbl_1 = (a == b)
    >>> comparison_qbl_1.qs.statevector()
    sqrt(2)*(|0>*|True>*|4>*|False> + |4>*|False>*|4>*|True>)/2

    The first tensor factor containing a boolean value is corresponding to
    ``comparison_qbl_0`` and the second one is ``comparison_qbl_1``.

    """

    def __init__(self, msize, exponent=0, qs=None, name=None, signed=False):
        # Boolean to indicate if the float is signed
        self.signed = signed

        # Exponent
        self.exponent = exponent

        # Size of the mantissa
        self.msize = msize

        # Array that consists of (log2(min), log2(max)) where min and max are the
        # minimal and maximal values of the absolutes that the QuantumFloat can
        # represent.
        self.mshape = np.array([exponent, exponent + msize])

        # Initialize QuantumVariable
        if signed:
            super().__init__(msize + 1, qs, name=name)
        else:
            super().__init__(msize, qs, name=name)

    # Define outcome_labels
    def decoder(self, i):
        if self.signed:
            res = signed_int_iso_inv(i, self.size - 1) * 2.0**self.exponent
        else:
            res = i * 2**self.exponent

        if self.exponent >= 0:
            return int(res)
        else:
            return res

    # def encoder(self, i):
    #     if self.signed:
    #         res = int(signed_int_iso(i/2**self.exponent, self.size-1))
    #     else:
    #         res =  int(i/2**self.exponent)

    #     return res

    def sb_poly(self, m=0):
        """
        Returns the semi-boolean polynomial of this `QuantumFloat` where `m` specifies
        the image extension parameter.

        For the technical details we refer to:
        https://ieeexplore.ieee.org/document/9815035


        Parameters
        ----------
        m : int, optional
            Image extension parameter. The default is 0.

        Returns
        -------
        Sympy expression
            The semi-boolean polynomial of this QuantumFloat.

        Examples
        --------

        >>> from qrisp import QuantumFloat
        >>> x = QuantumFloat(3, -1, signed = True, name = "x")
        >>> print(x.sb_poly(5))
        0.5*x_0 + 1.0*x_1 + 2.0*x_2 + 28.0*x_3

        """

        if m == 0:
            m = self.size

        symbols = [sp.symbols(self.name + "_" + str(i)) for i in range(self.size)]

        poly = sum([2.0 ** (i) * symbols[i] for i in range(self.size)])

        if self.signed:
            poly += (2.0 ** (m + 1) - 2.0 ** (self.size)) * symbols[-1]

        return 2**self.exponent * poly

    def encode(self, encoding_number, rounding=False, permit_dirtyness=False):
        if rounding:
            # Round value to closest fitting number
            outcome_labels = [self.decoder(i) for i in range(2**self.size)]
            encoding_number = outcome_labels[
                np.argmin(np.abs(encoding_number - np.array(outcome_labels)))
            ]

        super().encode(encoding_number, permit_dirtyness=permit_dirtyness)

    @gate_wrap(permeability="args", is_qfree=True)
    def __mul__(self, other):
        if isinstance(other, QuantumFloat):
            return q_mult(self, other)
        elif isinstance(other, int):
            bit_shift = 0
            while not other % 2:
                other = other >> 1
                bit_shift += 1

            if self.signed or other < 0:
                output_qf = QuantumFloat(
                    self.msize + int(np.ceil(np.log2(abs(other)))),
                    self.exponent,
                    signed=True,
                )
            else:
                output_qf = QuantumFloat(
                    self.msize + int(np.ceil(np.log2(abs(other)))),
                    self.exponent,
                    signed=False,
                )

            polynomial_encoder([self], output_qf, other * sp.Symbol("x"))

            output_qf << bit_shift

            return output_qf
        else:
            raise Exception(
                "QuantumFloat multiplication for type " + str(type(other)) + ""
                " not implemented (available are QuantumFloat and int)"
            )

    @gate_wrap(permeability="args", is_qfree=True)
    def __add__(self, other):
        if isinstance(other, QuantumFloat):
            return sbp_add(self, other)
        elif isinstance(other, (int, float)):
            res = self.duplicate(init=True)
            res += other
            return res
        else:
            raise Exception(
                "Addition with type " + str(type(other)) + " not implemented"
            )

    @gate_wrap(permeability="args", is_qfree=True)
    def __sub__(self, other):
        from qrisp.arithmetic import sbp_sub

        if isinstance(other, QuantumFloat):
            return sbp_sub(self, other)
        elif isinstance(other, (int, float)):
            res = self.duplicate(init=True)
            res -= other
            return res
        else:
            raise Exception(
                "Subtraction with type " + str(type(other)) + " not implemented"
            )

    __radd__ = __add__
    __rmul__ = __mul__

    @gate_wrap(permeability="args", is_qfree=True)
    def __rsub__(self, other):
        from qrisp import x
        from qrisp.arithmetic import sbp_sub

        if isinstance(other, QuantumFloat):
            return sbp_sub(other, self)
        elif isinstance(other, (int, float)):
            res = self.duplicate(init=True)
            if not res.signed:
                res.add_sign()
            x(res)
            res -= other - 2**res.exponent
            return res
        else:
            raise Exception(
                "Subtraction with type " + str(type(other)) + " not implemented"
            )

    @gate_wrap(permeability="args", is_qfree=True)
    def __truediv__(self, other):
        from qrisp.arithmetic import q_div

        return q_div(self, other)

    @gate_wrap(permeability="args", is_qfree=True)
    def __floordiv__(self, other):
        if self.signed or other.signed:
            raise Exception("Floor division not implemented for signed QuantumFloats")

        if self.exponent < 0 or other.exponent < 0:
            raise Exception(
                "Tried to perform floor division on non-integer QuantumFloats"
            )
        from qrisp.arithmetic import q_div

        return q_div(self, other, prec=0)

    @gate_wrap(permeability="args", is_qfree=True)
    def __pow__(self, power):
        if power != -1:
            raise Exception("Currently the only supported power is -1")

        from qrisp.arithmetic import qf_inversion

        return qf_inversion(self)

    @gate_wrap(permeability=[1], is_qfree=True)
    def __iadd__(self, other):
        if isinstance(other, QuantumFloat):
            input_qf_list = [other]
            poly = sp.symbols("x")

            polynomial_encoder(input_qf_list, self, poly)

        elif isinstance(other, (int, float)):
            # self.incr(other)

            if not int(other / 2**self.exponent) == other / 2**self.exponent:
                raise Exception(
                    "Tried to perform in-place addition with invalid number. "
                    "QuantumFloat precision too low."
                )

            input_qf_list = []
            poly = sp.sympify(other)

            polynomial_encoder(input_qf_list, self, poly)

        else:
            raise Exception(
                "In-place addition for type " + str(type(other)) + " not implemented"
            )

        return self

    @gate_wrap(permeability=[1], is_qfree=True)
    def __isub__(self, other):
        if isinstance(other, QuantumFloat):
            input_qf_list = [other]
            poly = -sp.symbols("x")

            polynomial_encoder(input_qf_list, self, poly)

        elif isinstance(other, (int, float)):
            if not int(other / 2**self.exponent) == other / 2**self.exponent:
                raise Exception(
                    "Tried to perform in-place subtraction with invalid number. "
                    "QuantumFloat precision too low."
                )

            input_qf_list = []
            poly = -sp.sympify(other)

            polynomial_encoder(input_qf_list, self, poly)

        else:
            raise Exception(
                "In-place substraction for type "
                + str(type(other))
                + " not implemented"
            )

        return self

    @gate_wrap(permeability=[], is_qfree=True)
    def __imul__(self, other):
        inpl_mult(self, other)

        return self

    def __rshift__(self, k):
        self.exp_shift(-k)
        return self

    def __lshift__(self, k):
        self.exp_shift(k)
        return self

    def __lt__(self, other):
        if not isinstance(other, (QuantumFloat, int, float)):
            raise Exception(f"Comparison with type {type(other)} not implemented")

        return lt(self, other)

    def __gt__(self, other):
        if not isinstance(other, (QuantumFloat, int, float)):
            raise Exception(f"Comparison with type {type(other)} not implemented")

        return gt(self, other)

    def __le__(self, other):
        if not isinstance(other, (QuantumFloat, int, float)):
            raise Exception(f"Comparison with type {type(other)} not implemented")

        return leq(self, other)

    def __ge__(self, other):
        if not isinstance(other, (QuantumFloat, int, float)):
            raise Exception(f"Comparison with type {type(other)} not implemented")

        return geq(self, other)

    def __eq__(self, other):
        if not isinstance(other, (QuantumFloat, int, float)):
            raise Exception(f"Comparison with type {type(other)} not implemented")

        return eq(self, other)

    def __ne__(self, other):
        if not isinstance(other, (QuantumFloat, int, float)):
            raise Exception(f"Comparison with type {type(other)} not implemented")

        return neq(self, other)

    def exp_shift(self, shift):
        """
        Performs an internal bit shift. Note that this method doesn't cost any
        quantum gates. For the quantum version of this method, see
        :meth:`quantum_bit_shift<qrisp.QuantumFloat.quantum_bitshift>`.

        Parameters
        ----------
        shift : int
            The amount to shift.

        Raises
        ------
        Exception
            Tried to shift QuantumFloat exponent by non-integer value

        Examples
        --------
        
        We create a QuantumFloat and perform a bitshift:
            
        >>> from qrisp import QuantumFloat
        >>> a = QuantumFloat(4)
        >>> a[:] = 2
        >>> a.exp_shift(2)
        >>> print(a)
        {8: 1.0}
        >>> print(a.qs)
        QuantumCircuit:
        ---------------
        a.0: ─────
             ┌───┐
        a.1: ┤ X ├
             └───┘
        a.2: ─────
        <BLANKLINE>
        a.3: ─────
        <BLANKLINE>          
        Live QuantumVariables:
        ----------------------
        QuantumFloat a
        
        """
        if not isinstance(shift, int):
            raise Exception("Tried to shift QuantumFloat exponent by non-integer value")

        self.exponent += shift
        self.mshape = self.mshape + shift

    def reduce(self, qubits, verify=False):
        QuantumVariable.reduce(self, qubits, verify)

        try:
            self.mshape[1] -= len(qubits)
            self.msize -= len(qubits)
        except TypeError:
            self.mshape[1] -= 1
            self.msize -= 1

    def extend(self, amount, position=-1):
        QuantumVariable.extend(self, amount, position=position)

        self.mshape[1] += amount
        self.msize += amount

    def add_sign(self):
        """
        Turns an unsigned QuantumFloat into its signed version.

        Raises
        ------
        Exception
            Tried to add sign to signed QuantumFloat.

        Examples
        --------

        >>> from qrisp import QuantumFloat
        >>> qf = QuantumFloat(4)
        >>> qf.signed
        False
        >>> qf.add_sign()
        >>> qf.signed
        True

        """

        if self.signed:
            raise Exception(r'Tried to add sign to signed QuantumFloat')

        self.extend(1, self.size - 1)
        self.mshape[1] -= 1
        self.msize -= 1
        self.signed = True

    def sign(self):
        r"""
        Returns the sign qubit.

        This qubit is in state $\ket{1}$ if the QuantumFloat holds a negative value and
        in state $\ket{0}$ otherwise.

        For more information about the encoding of negative numbers check the
        `publication <https://ieeexplore.ieee.org/document/9815035>`_.

        .. warning::

            Performing an X gate on this qubit does not flip the sign! Use inplace
            multiplication instead.

            >>> from qrisp import QuantumFloat
            >>> qf = QuantumFloat(3, signed = True)
            >>> qf[:] = 3
            >>> qf *= -1
            >>> print(qf)
            {-3: 1.0}

        Raises
        ------
        Exception
            Tried to retrieve sign qubit of unsigned QuantumFloat.

        Returns
        -------
        Qubit
            The qubit holding the sign.

        Examples
        --------

        We create a QuantumFloat, initiate a state that has probability 2/3 of being
        negative and entangle a QuantumBool with the sign qubit.

        >>> from qrisp import QuantumFloat, QuantumBool, cx
        >>> qf = QuantumFloat(4, signed = True)
        >>> n_amp = 1/3**0.5
        >>> qf[:] = {-1 : n_amp, -2 : n_amp, 1 : n_amp}
        >>> qbl = QuantumBool()
        >>> cx(qf.sign(), qbl)
        >>> print(qbl)
        {True: 0.6667, False: 0.3333}

        """
        if not self.signed:
            raise Exception("Tried to retrieve sign qubit of unsigned QuantumFloat")

        return self[-1]

    def init_from(
        self, other, ignore_rounding_errors=False, ignore_overflow_errors=False
    ):
        copy_qf(
            self,
            other,
            ignore_rounding_errors=ignore_rounding_errors,
            ignore_overflow_errors=ignore_overflow_errors,
        )

    def incr(self, x=None):
        from qrisp.arithmetic.incrementation import increment

        if x is None:
            x = 2**self.exponent
        increment(self, x)

    def __hash__(self):
        return id(self)

    def significant(self, k):
        """
        Returns the qubit with significance $k$.

        Parameters
        ----------
        k : int
            The significance.

        Raises
        ------
        Exception
            Tried to retrieve invalid significant from QuantumFloat

        Returns
        -------
        Qubit
            The Qubit with significance $k$.

        Examples
        --------

        We create a QuantumFloat and flip a qubit of specified significance.

        >>> from qrisp import QuantumFloat, x
        >>> qf = QuantumFloat(6, -3)
        >>> x(qf.significant(-2))
        >>> print(qf)
        {0.25: 1.0}

        The qubit with significance $-2$ corresponds to the value $0.25 = 2^{-2}$.

        >>> x(qf.significant(2))
        {4.25: 1.0}

        The qubit with significance $2$ corresponds to the value $4 = 2^{2}$.

        """

        sig_list = list(range(self.mshape[0], self.mshape[1]))

        if k not in sig_list:
            raise Exception(
                f"Tried to retrieve invalid significant {k} "
                f"from QuantumFloat with mantissa shape {self.mshape}"
            )

        return self[sig_list.index(k)]

    def truncate(self, x):
        """
        Receives a regular float and returns the float that is closest to the input but
        can still be encoded.

        Parameters
        ----------
        x : float
            A float that is supposed to be truncated.

        Returns
        -------
        float
            The truncated float.

        Examples
        --------

        We create a QuantumFloat and round a value to fit the encoder and subsequently
        initiate:

        >>> from qrisp import QuantumFloat
        >>> qf = QuantumFloat(4, -1)
        >>> value = 0.5102341
        >>> qf[:] = value
        Exception: Value 0.5102341 not supported by encoder.
        >>> rounded_value = qf.truncate(value)
        >>> rounded_value
        0.5
        >>> qf[:] = rounded_value
        >>> print(qf)
        {0.5: 1.0}

        """
        decoder_values = np.array([self.decoder(i) for i in range(2**self.size)])

        return decoder_values[np.argmin(np.abs(decoder_values - x))]
    
    def get_ev(self, **mes_kwargs):
        """
        Retrieves the expectation value of self.

        Parameters
        ----------
        **mes_kwargs : dict
            Keyword arguments for the measurement. See :meth:`qrisp.QuantumVariable.get_measurement` for more information.

        Returns
        -------
        float
            The expectation value.
            
        Examples
        --------
        
        We set up a QuantumFloat in uniform superposition and retrieve the expectation value:
            
        >>> from qrisp import QuantumFloat, h
        >>> qf = QuantumFloat(4)
        >>> h(qf)
        >>> qf.get_ev()
        7.5
        
        """
        
        mes_res = self.get_measurement(**mes_kwargs)
        
        return sum([k*v for k,v in mes_res.items()])
    
    def quantum_bit_shift(self, shift_amount):
        """
        Performs a bit shift in the quantum device.
        While :meth:`exp_shift<qrisp.QuantumFloat.exp_shift>` performs a bit shift
        in the compiler (thus costing no quantum gates) this method performs the
        bitshift on the hardware.
        
        This has the advantage, that it can be controlled if called within a 
        :ref:`ControlEnvironment` and furthermore admits bit shifts based on the
        state of a QuantumFloat
        
        .. note::
            
            Bit bit shifts based on a QuantumFloat are currently only possible
            if both self and ``shift_amount`` are unsigned.
            
        .. warning::
            
            Quantum bit shifting extends the QuantumFloat (ie. it allocates 
            additional qubits).

        Parameters
        ----------
        shift_amount : int or QuantumFloat
            The amount to shift.
        
        Raises
        ------
        Exception
            Tried to shift QuantumFloat exponent by non-integer value
        Exception
            Quantum-quantum bitshifting is currently only supported for unsigned arguments
        
        Examples
        --------
        
        We create a QuantumFloat and a QuantumBool to perform a controlled bit shift.
        
        ::
            
            from qrisp import QuantumFloat, QuantumBool, h
            qf = QuantumFloat(4)
            qf[:] = 1
            qbl = QuantumBool()
            h(qbl)

            with qbl:
                qf.quantum_bit_shift(2)
        
        Evaluate the result
        
        >>> print(qf.qs.statevector())
        sqrt(2)*(|1>*|False> + |4>*|True>)/2

        """
        quantum_bit_shift(self, shift_amount)
        
        


def create_output_qf(operands, op):
    if isinstance(op, sp.core.expr.Expr):
        from qrisp.arithmetic.poly_tools import expr_to_list

        expr_list = expr_to_list(op)

        for i in range(len(expr_list)):
            if not isinstance(expr_list[i][0], sp.Symbol):
                expr_list[i].pop(0)

        operands.sort(key=lambda x: x.name)

        def prod(iter):
            iter = list(iter)
            a = iter[0]
            for i in range(1, len(iter)):
                a *= iter[i]

            return a

        from sympy import Abs, Poly, Symbol

        poly = Poly(op)
        monom_list = [
            a * prod(x**k for x, k in zip(poly.gens, mon))
            for a, mon in zip(poly.coeffs(), poly.monoms())
        ]

        max_value_dic = {Symbol(qf.name): 2.0 ** qf.mshape[1] for qf in operands}
        min_value_dic = {Symbol(qf.name): 2.0 ** qf.mshape[0] for qf in operands}

        abs_poly = sum([Abs(monom) for monom in monom_list], 0)

        min_poly_value = min(
            [float(Abs(monom).subs(min_value_dic)) for monom in monom_list]
        )

        max_poly_value = float(abs_poly.subs(max_value_dic))

        min_sig = int(np.floor(np.log2(min_poly_value)))
        max_sig = int(np.ceil(np.log2(max_poly_value)))

        msize = max_sig - min_sig
        exponent = min_sig

        signed = bool(sum([int(operand.signed) for operand in operands]))

        return QuantumFloat(msize, exponent=exponent, signed=signed)

    if op == "add":
        signed = operands[0].signed or operands[1].signed
        exponent = min(operands[0].exponent, operands[1].exponent)

        max_sig = int(
            np.ceil(
                np.log2(int(2 ** operands[0].mshape[1] + 2 ** operands[1].mshape[1]))
            )
        )
        msize = max_sig - exponent + 1
        return QuantumFloat(
            msize, exponent, operands[0].qs, signed=signed, name="add_res*"
        )

    if op == "mul":
        signed = operands[0].signed or operands[1].signed

        if operands[0].reg == operands[1].reg and (
            operands[0].signed and operands[1].signed
        ):
            signed = False

        return QuantumFloat(
            operands[0].msize
            + operands[1].msize
            + operands[0].signed * operands[1].signed,
            operands[0].exponent + operands[1].exponent,
            operands[0].qs,
            signed=signed,
            name="mul_res*",
        )

    if op == "sub":
        exponent = min(operands[0].exponent, operands[1].exponent)
        max_sig = int(
            np.ceil(
                np.log2(int(2 ** operands[0].mshape[1] + 2 ** operands[1].mshape[1]))
            )
        )
        msize = max_sig - exponent + 1

        return QuantumFloat(
            msize, exponent, operands[0].qs, signed=True, name="sub_res*"
        )


# Initiates the value of qf2 into qf1 where qf1 has to hold the value 0
def copy_qf(qf1, qf2, ignore_overflow_errors=False, ignore_rounding_errors=False):
    # Lists that translate Qubit index => Significance
    qf1_sign_list = [qf1.exponent + i for i in range(qf1.size)]
    qf2_sign_list = [qf2.exponent + i for i in range(qf2.size)]

    # Check overflow/underflow
    if max(qf1_sign_list) < max(qf2_sign_list) and not ignore_overflow_errors:
        raise Exception(
            "Copy operation would result in overflow "
            "(use ignore_overflow_errors = True)"
        )

    if min(qf1_sign_list) > min(qf2_sign_list) and not ignore_rounding_errors:
        raise Exception(
            "Copy operation would result in rounding "
            "(use ignore_rounding_errors = True)"
        )

    qs = qf1.qs

    if qf2.signed:
        if not qf1.signed:
            raise Exception("Tried to copy signed into unsigend float")

        # Remove last entry from significance list (last qubit is the sign qubit)
        qf2_sign_list.pop(-1)
        qf1_sign_list.pop(-1)

    for i in range(len(qf1_sign_list)):
        # If we are in a realm where both floats have overlapping significance
        # => CNOT into each other
        if qf1_sign_list[i] in qf2_sign_list:
            qf2_index = qf2_sign_list.index(qf1_sign_list[i])
            qs.cx(qf2[qf2_index], qf1[i])
            continue

        # Otherwise copy the sign bit into the bits of higher significance than qf2
        if qf1_sign_list[i] > max(qf2_sign_list) and qf2.signed:
            qs.cx(qf2[-1], qf1[i])

    # Copy the sign bit
    if qf2.signed:
        qs.cx(qf2[-1], qf1[-1])



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


import numpy as np

from qrisp.core.quantum_array import QuantumArray
from qrisp.qtypes.quantum_char import QuantumChar

nisq_init_quantum_char = QuantumChar(name="nisq_q_char")
init_quantum_char = QuantumChar(nisq_char=False, name="q_char")


class QuantumString(QuantumArray):
    """
    The QuantumString is the quantum equivalent of a string. It is implemented as a
    :ref:`QuantumArray` of :ref:`QuantumChars <QuantumChar>`.

    >>> from qrisp import QuantumString
    >>> q_str = QuantumString(size = len("hello world"))
    >>> q_str[:] = "hello world"
    >>> print(q_str)
    {'hello world': 1.0}

    It is also possible to have a QuantumString containing non-nisq chars

    >>> q_str_nn = QuantumString(size = len("hello world"), nisq_char = False)
    >>> q_str_nn[:] = "hello world"
    >>> print(q_str_nn)
    {'hello world': 1.0}

    This requires however considerably more qubits

    >>> print(len(q_str.qs.qubits))
    55
    >>> print(len(q_str_nn.qs.qubits))
    88

    Similar to its parent class, the size of a QuantumString does not have to be
    specified at creation

    >>> q_str = QuantumString()
    >>> q_str[:] = "hello world"
    >>> print(q_str)
    {'hello world': 1.0}

    **Concatenation**

    QuantumStrings provide a number of methods to concatenate:

    >>> q_str_0 = QuantumString()
    >>> q_str_1 = QuantumString()
    >>> q_str_2 = QuantumString()
    >>> q_str_0[:] = "hello"
    >>> q_str_1 += " "
    >>> q_str_2[:] = "world"
    >>> q_str_3 = q_str_1 + q_str_2
    >>> q_str_0 += q_str_3
    >>> print(q_str_0)
    {'hello world': 1.0}

    Note that these QuantumStrings share memory - i.e. if we modify a QuantumChar in one
    of them, this will potentially affect the others:

    >>> from qrisp import h
    >>> h(q_str_2[0][0])
    >>> print(q_str_0)
    {'hello world': 0.5, 'hello xorld': 0.5}

    """

    def __new__(subtype, size=0, qs=None, nisq_char=True):
        if nisq_char:
            return QuantumArray.__new__(subtype, nisq_init_quantum_char, size, qs=qs)
        else:
            return QuantumArray.__new__(subtype, init_quantum_char, size, qs=qs)

    def get_measurement(self, **kwargs):
        mes_result = QuantumArray.get_measurement(self, **kwargs)

        return_dic = {}

        for k, v in mes_result.items():
            return_dic["".join(list(k))] = v

        return return_dic

    def __add__(self, other):
        return np.concatenate((self, other)).view(QuantumString)

    def decoder(self, code_int):
        res_array = QuantumArray.decoder(self, code_int)
        res_str = ""

        for i in range(len(res_array)):
            res_str += res_array[i]

        return res_str

    def encoder(self, encoding_str):
        encoding_array = np.zeros(len(encoding_str), dtype="object")

        for i in range(len(encoding_array)):
            encoding_array[i] = encoding_str[i]

        return QuantumArray.encoder(self, encoding_array)

    def __setitem__(self, key, value):
        if isinstance(key, int):
            QuantumArray.__setitem__(self, key, value)
            return

        QuantumArray.__setitem__(self, key, [ch for ch in value])

    def __iadd__(self, other):
        from qrisp import merge

        if isinstance(other, QuantumString):
            self.resize((len(self) + len(other),), refcheck=False)
            merge(self.qs, other.qs)
            for i in range(len(other)):
                np.ndarray.__setitem__(self, i + len(self) - len(other), other[i])

        elif isinstance(other, str):
            if self.shape_specified:
                self.resize((len(self) + len(other),), refcheck=False)
            else:
                self.set_shape(len(other))
            for i in range(len(other)):
                np.ndarray.__setitem__(
                    self, i + len(self) - len(other), self.qtype.duplicate(qs=self.qs)
                )
                self[i + len(self) - len(other)] = other[i]

        elif isinstance(other, QuantumChar):
            self.resize((len(self) + 1,), refcheck=False)
            np.ndarray.__setitem__(self, len(self) - 1, other)

        return self



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


class Clbit:
    """
    This class represents classical bits. Classical bits are created by supplying the
    identifier string.

    Examples
    --------

    We create a Clbit and add it to a QuantumCircuit. After applying an Hadamard-Gate,
    we measure into the Clbit.

    >>> from qrisp import Clbit, QuantumCircuit
    >>> qc = QuantumCircuit(1)
    >>> qc.h(0)
    >>> mes_result = Clbit("mes_result")
    >>> qc.add_clbit(mes_result)
    >>> qc.measure(0, mes_result)
    >>> qc.run()
    {'0': 5000, '1': 5000}

    """

    def __init__(self, identifier):
        self.identifier = identifier

    def __str__(self):
        return self.identifier

    def __repr__(self):
        return self.identifier

    def __hash__(self):
        return hash(self.identifier)



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


import os
import threading

import connexion
import six
from connexion.apps.flask_app import FlaskJSONEncoder
from openapi_server.models.backend_status import BackendStatus
from openapi_server.models.base_model_ import Model
from openapi_server.models.inline_object import InlineObject  # noqa: E501

import qrisp.interface.openapi_interface.codegen.server.openapi_server as openapi_server


class JSONEncoder(FlaskJSONEncoder):
    include_nulls = False

    def default(self, o):
        if isinstance(o, Model):
            dikt = {}
            for attr, _ in six.iteritems(o.openapi_types):
                value = getattr(o, attr)
                if value is None and not self.include_nulls:
                    continue
                attr = o.attribute_map[attr]
                dikt[attr] = value
            return dikt
        return FlaskJSONEncoder.default(self, o)


# Returns the hosts ip
def get_ip():
    import socket

    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.settimeout(0)
    try:
        # doesn't even have to be reachable
        s.connect(("10.255.255.255", 1))
        IP = s.getsockname()[0]
    except Exception:
        IP = "127.0.0.1"
    finally:
        s.close()
    return IP


# This class describes a Backend Server
class BackendServer:
    # For the constructor we specify a function that runs the circuit, a socket ip
    # adress, a port and the name.
    # Note that the name is supposed to be unique, that is if this python instance
    # already runs a threat with that name, no new server is started
    def __init__(
        self,
        run_func,
        socket_ip_address=None,
        port=None,
        name="generic_quantum_backend_server",
        ping_func=None,
        online_status=True,
    ):
        self.app = connexion.App(
            __name__, specification_dir="./codegen/server/openapi_server/openapi/"
        )
        self.app.app.json_encoder = JSONEncoder

        @self.app.route("/run", methods=["POST"])
        def run():
            inline_object = InlineObject.from_dict(connexion.request.get_json())
            qc = inline_object.qc
            shots = inline_object.shots
            token = inline_object.token

            return run_func(qc, shots, token)

        if ping_func is not None:

            @self.app.route("/ping", methods=["GET"])
            def ping():
                res = ping_func()
                res.online = pass_online_status_bool_by_reference()
                return res.to_dict()

        self.online_status = online_status
        pass_online_status_bool_by_reference = lambda: self.online_status

        self.name = name

        if port is None:
            port = 9010

        if socket_ip_address is None:
            socket_ip_address = get_ip()

        self.socket_ip_address = socket_ip_address
        self.port = port

    # Starts the server
    def start(self):
        # self.app.run(port=self.port)

        # import logging
        # log = logging.getLogger('werkzeug')
        # log.setLevel(logging.ERROR)
        # Create thread
        from waitress import serve

        def wrapper():
            serve(self.app, host=self.socket_ip_address, port=self.port)

        thr = threading.Thread(target=wrapper)
        thr.setDaemon(True)

        # Start the thread
        thr.start()



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

# Created by ann81984 at 28.06.2022
from qrisp import (
    QuantumDictionary,
    QuantumFloat,
    custom_qv,
    h,
    multi_measurement,
    QuantumVariable,
)

# This example demonstrates the use of a QuantumDictionary with flexible types


def test_quantum_dictionary():
    # Create a QuantumDictionary
    test_qd = QuantumDictionary()

    # Fill in some entries (works precisely like normal Python dictionaries)
    test_qd["hello"] = "hallo"
    test_qd["world"] = "welt"

    # Create Key QuantumVariable (the constructor takes a
    # list of values, which the QuantumVariable is supposed to represent)
    key_qv = custom_qv(["hello", "world"])

    # Bring into superposition
    h(key_qv)

    print(key_qv)
    assert str(key_qv) == "{'hello': 0.5, 'world': 0.5}"
    # Yields {'hello': 0.5, 'world': 0.5}

    # Use QuantumDictionary to "dereference" the values into a new QuantumVariable
    value_qv = test_qd[key_qv]

    # Evaluate values
    print(value_qv)
    assert str(value_qv) == "{'hallo': 0.5, 'welt': 0.5}"
    # Yields {'hallo': 0.5, 'welt': 0.5}

    # Now we demonstrate how a QuantumDictionary with a specified type can be created
    # The return values from dereferencing this QuantumDictionary are now of the type QuantumFloat(4, -1)
    float_qd = QuantumDictionary(return_type=QuantumFloat(4, -1))

    # Encode some values (performed as with usual Python dictionaries)
    float_qd.update({"hallo": 1.5, "welt": 0.5})

    # Dereference
    value_qf = float_qd[value_qv]

    # Results are indeed QuantumFloat
    print(type(value_qf))
    assert type(value_qf).__name__ == "QuantumFloat"

    # Usual arithmetic is possible
    value_qf += 2.5

    # Evaluate results
    assert value_qf.get_measurement() == {3.0: 0.5, 4.0: 0.5}
    # Yields {4.0: 0.5, 3.0: 0.5}

    test_qd[(1, 2)] = 73
    test_qd[(0, 2)] = 37
    qf1 = QuantumFloat(1)
    qf2 = QuantumFloat(2)
    h(qf1)
    qf2[:] = 2
    res = test_qd[(qf1, qf2)]
    assert multi_measurement([qf1, qf2, res]) == {(0, 2, 37): 0.5, (1, 2, 73): 0.5}

    qtype = QuantumFloat(4, -2, signed=True)

    float_qd = QuantumDictionary(return_type=qtype)

    float_qd["hello"] = 0.5
    float_qd["world"] = -1
    key_qv = QuantumVariable.custom([1, 42, "hello", "world"])
    key_qv[:] = {"hello": 1, "world": 1}

    qf = qtype.duplicate()
    float_qd.load(key_qv, qf)

    qf2 = qtype.duplicate()
    float_qd.load(key_qv, qf2, synth_method="pprm")

    assert qf2.get_measurement() == {0.5: 0.5, -1.0: 0.5}



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


import copy
import weakref

import matplotlib.pyplot as plt
import numpy as np

from qrisp.core.compilation import qompiler


class QuantumVariable:
    """
    The QuantumVariable is the quantum equivalent of a regular variable in classical
    programming languages. All :ref:`quantum types <QuantumTypes>` inherit from this
    class. The QuantumVariable allows many automizations and quality of life
    improvements such as hidden qubit management, de/encoding to human readable labels
    or typing.

    Each QuantumVariable is registered in a :ref:`QuantumSession`. It can be accessed
    using the ``.qs`` attribute:

    >>> from qrisp import QuantumVariable
    >>> example_qv = QuantumVariable(3)
    >>> quantum_session = example_qv.qs

    The qubits of the QuantumVariable are stored as a list in the ``.reg`` attribute

    >>> qubits = example_qv.reg

    To quickly access the qubits of a given variable, we use the [ ] operator:

    >>> qubit_2 = example_qv[2]

    We can find out about the amount of qubits in the QuantumVariable with the ``.size``
    attribute

    >>> example_qv.size
    3

    **Naming**

    QuantumVariables can be given names to identify them independently of their naming
    as Python objects.

    >>> example_qv_2 = QuantumVariable(3, name = "alice")
    >>> example_qv_2.name
    'alice'

    If not explicitely specified during construction, a name is determined
    automatically. Qrisp will try to infer the name of the Python variable and if that
    fails, a generic name is given.

    >>> example_qv.name
    'example_qv'

    In order to keep the generated quantum circuits comprehensive, the qubits are named
    after their containing QuantumVariable with an extra number, which indicates their
    index.

    >>> from qrisp import cx
    >>> cx(example_qv, example_qv_2)
    >>> print(example_qv.qs)
    QuantumCircuit:
    ---------------
    example_qv.0: ──■────────────
                    │
    example_qv.1: ──┼────■───────
                    │    │
    example_qv.2: ──┼────┼────■──
                  ┌─┴─┐  │    │
         alice.0: ┤ X ├──┼────┼──
                  └───┘┌─┴─┐  │
         alice.1: ─────┤ X ├──┼──
                       └───┘┌─┴─┐
         alice.2: ──────────┤ X ├
                            └───┘
    Live QuantumVariables:
    ----------------------
    QuantumVariable example_qv
    QuantumVariable alice

    QuantumSessions can only contain uniquely named QuantumVariables. If two
    QuantumSessions are :ref:`merged <SessionMerging>` containing identically named
    QuantumVariables, the more recently created QuantumVariable will be renamed:

    ::

        from qrisp import QuantumFloat

        s = QuantumFloat(5)

        for i in range(4):
            temp = QuantumFloat(4)
            temp[:] = 2**i
            s += temp

    >>> print(s.qs)
    QuantumCircuit:
    ---------------
                   ┌───────────┐┌───────────┐┌───────────┐┌───────────┐
         s.0: ─────┤0          ├┤0          ├┤0          ├┤0          ├
                   │           ││           ││           ││           │
         s.1: ─────┤1          ├┤1          ├┤1          ├┤1          ├
                   │           ││           ││           ││           │
         s.2: ─────┤2          ├┤2          ├┤2          ├┤2          ├
                   │           ││           ││           ││           │
         s.3: ─────┤3          ├┤3          ├┤3          ├┤3          ├
                   │           ││           ││           ││           │
         s.4: ─────┤4 __iadd__ ├┤4          ├┤4          ├┤4          ├
              ┌───┐│           ││           ││           ││           │
      temp.0: ┤ X ├┤5          ├┤           ├┤           ├┤           ├
              └───┘│           ││           ││           ││           │
      temp.1: ─────┤6          ├┤  __iadd__ ├┤           ├┤           ├
                   │           ││           ││           ││           │
      temp.2: ─────┤7          ├┤           ├┤           ├┤           ├
                   │           ││           ││           ││           │
      temp.3: ─────┤8          ├┤           ├┤  __iadd__ ├┤           ├
                   └───────────┘│           ││           ││           │
    temp_1.0: ──────────────────┤5          ├┤           ├┤           ├
              ┌───┐             │           ││           ││           │
    temp_1.1: ┤ X ├─────────────┤6          ├┤           ├┤  __iadd__ ├
              └───┘             │           ││           ││           │
    temp_1.2: ──────────────────┤7          ├┤           ├┤           ├
                                │           ││           ││           │
    temp_1.3: ──────────────────┤8          ├┤           ├┤           ├
                                └───────────┘│           ││           │
    temp_2.0: ───────────────────────────────┤5          ├┤           ├
                                             │           ││           │
    temp_2.1: ───────────────────────────────┤6          ├┤           ├
              ┌───┐                          │           ││           │
    temp_2.2: ┤ X ├──────────────────────────┤7          ├┤           ├
              └───┘                          │           ││           │
    temp_2.3: ───────────────────────────────┤8          ├┤           ├
                                             └───────────┘│           │
    temp_3.0: ────────────────────────────────────────────┤5          ├
                                                          │           │
    temp_3.1: ────────────────────────────────────────────┤6          ├
                                                          │           │
    temp_3.2: ────────────────────────────────────────────┤7          ├
              ┌───┐                                       │           │
    temp_3.3: ┤ X ├───────────────────────────────────────┤8          ├
              └───┘                                       └───────────┘
    Live QuantumVariables:
    ----------------------
    QuantumFloat s
    QuantumFloat temp
    QuantumFloat temp_1
    QuantumFloat temp_2
    QuantumFloat temp_3

    Renaming does not happen for names given through the ``name`` keyword, unless the
    name ends with a ``*``.

    >>> example_qv_3 = QuantumVariable(3, name = "alice")
    >>> cx(example_qv, example_qv_3)
    Exception: Tried to merge QuantumSession containing identically named
    QuantumVariables
    >>> example_qv_4 = QuantumVariable(3, name = "alice*")
    >>> cx(example_qv, example_qv_4)
    >>> example_qv_4.name
    'alice_1'


    Examples
    --------

    Writing a function that brings an arbitrary QuantumVariable into a GHZ state

    ::

        from qrisp import QuantumVariable, h, cx
        def GHZ(qv):
             h(qv[0])
             for i in range(1, qv.size):
                 cx(qv[0], qv[i])

    Evaluation:

    >>> qv = QuantumVariable(5)
    >>> GHZ(qv)
    >>> print(qv)
    {'00000': 0.5, '11111': 0.5}


    """

    live_qvs = []
    creation_counter = np.zeros(1)
    name_tracker = {}

    def __init__(self, size, qs=None, name=None):
        r"""
        Constructs a QuantumVariable - possibly with a given name or in a given
        QuantumSession.

        Parameters
        ----------
        size : int
            The amount of qubits this QuantumVariable contains.
        qs : QuantumSession, optional
            A QuantumSession object, where the QuantumVariable is supposed to be
            registered. The default is None.
        name : string, optional
            A name which uniquely identifies the QuantumVariable. If ended with a
            \*, name is allowed to be updated if a naming collision arises. By default,
            Qrisp will try to infer the name of the Python variable - otherwise a
            generic name is given.

        """

        if size < 0:
            raise Exception(
                f"Tried to create QuantumVariable with invalid qubit amount {size}"
            )

        # Store quantum session
        from qrisp.core import QuantumSession, merge_sessions

        self.qs = QuantumSession()

        if qs is not None:
            merge_sessions(self.qs, qs)

        self.size = int(size)

        self.user_given_name = False
        # If name is given, register variable in session manager
        if name is not None:
            self.user_given_name = True

            if name[-1] == "*":
                name = name[:-1]
                self.user_given_name = False

                try:
                    self.name = name
                    self.qs.register_qv(self)

                except RuntimeError:
                    i = 0
                    while True:
                        try:
                            self.name = name + "_" + str(i)
                            self.qs.register_qv(self)
                        except RuntimeError:
                            i += 1
                            continue
                        break

            else:
                self.name = name
                self.qs.register_qv(self)

        # Otherwise try to infer from code inspection
        else:
            from qrisp.misc import find_calling_line

            if type(self) is QuantumVariable:
                line = find_calling_line(1)
            else:
                line = find_calling_line(2)
            split_line = line.split("=")
            name_found = False

            if len(split_line) >= 2:
                python_var_name = split_line[0]

                if split_line[1].replace(" ", "")[:7] == "Quantum":
                    python_var_name = python_var_name.split(" ")[0]
                    python_var_name = python_var_name.split(" ")[-1]

                    # name = self.get_unique_name(python_var_name)
                    name = python_var_name

                    self.name = name
                    self.qs.register_qv(self)
                    name_found = True

            # If this didn't work, generate a generic, unique name
            if not name_found:
                while True:
                    try:
                        self.name = self.get_unique_name()
                        self.qs.register_qv(self)
                        break
                    except RuntimeError:
                        pass

        import weakref

        # This attribute tracks the created QuantumVariables for the
        # auto_uncompute decorator
        # We use weak references as some qrisp modules rely on reference counting
        QuantumVariable.live_qvs.append(weakref.ref(self))
        self.creation_time = int(self.creation_counter[0])
        self.creation_counter += 1

    def delete(self, verify=False, recompute=False):
        r"""
        This method is for deleting a QuantumVariable and thus freeing up and resetting
        the used qubits.

        Note that this method has a different function than the destructor. Calling this
        method will tell the QuantumSession to mark the used qubits as free and apply a
        reset gate.
        If set to True, the keyword verify will cause a simulation to check, wether the
        deleted qubits are in the $\ket{0}$ state prior to resetting. This is helpfull
        during debugging, as it indicates wether the uncomputation of this
        QuantumVariable was successfull.

        After deletion, the QuantumVariable object is basically unchanged but an error
        will be raised if further operations on the deleted qubits are attempted.

        Parameters
        ----------
        verify : bool, optional
            If this bool is set to True, Qrisp will verify that the deleted qubits are
            indeed in the $\ket{0}$ state. The default is ``False``.
        recompute : bool, optional
            If set to ``True``, this QuantumVariable can be recomputed if it is required
            for the uncomputation of another QuantumVariable. For more information on
            the (dis)advantages check :ref:`recomputation <recomputation>`. The default
            is ``False``.

        Raises
        ------
        Exception
            Tried to delete qubits not in \|0> state.


        Examples
        --------

        We create a QuantumVariable, execute some gates and try to delete with
        verify = True

        >>> from qrisp import QuantumVariable, x, h
        >>> qv = QuantumVariable(2)
        >>> x(qv[0])
        >>> h(qv[1])
        >>> qv.delete(verify = True)
        Exception: Tried to delete qubits not in |0> state.

        We now (manually) uncompute the gates

        >>> x(qv[0])
        >>> h(qv[1])
        >>> qv.delete(verify = True)
        >>> qv.is_deleted()
        True
        >>> x(qv[0])
        Exception: Tried to perform operation x on unallocated qubit qv_1.0.

        """

        if self.is_deleted():
            return

        self.qs.delete_qv(self, verify)

        i = 0
        while i < len(QuantumVariable.live_qvs):
            if QuantumVariable.live_qvs[i]() is None:
                QuantumVariable.live_qvs.pop(i)
                continue

            if QuantumVariable.live_qvs[i]().name == self.name:
                QuantumVariable.live_qvs.pop(i)
                break

            i += 1

        if recompute:
            for qb in self.reg:
                qb.recompute = True

    def is_deleted(self):
        for qb in self.reg:
            if qb.allocated:
                return False
        else:
            return True

    def duplicate(self, name=None, qs=None, init=False):
        r"""
        Duplicates the QuantumVariable in the sense that a new QuantumVariable is
        created with same type and parameters but initialized in the $\ket{0}$ state.

        Parameters
        ----------
        name : string, optional
            A unique name to identify that QuantumVariable. If not given, a name will be
            generated.
        qs : QuantumSession, optional
            A QuantumSession, where the result should be registered. If not given, a new
            QuantumSession will be generated.
        init : bool, optional
            If set to True, the :meth:`init_from <qrisp.QuantumVariable.init_from>`
            method of the result will be called on self. The default is False.

        Returns
        -------
        duplicate : Type of self
            The duplicated QuantumVariable.

        Examples
        --------

        We create a QuantumFloat and duplicate:

        >>> from qrisp import QuantumFloat
        >>> qf_0 = QuantumFloat(4, signed = False)
        >>> qf_1 = qf_0.duplicate()
        >>> type(qf_1)
        qrisp.qtypes.quantum_float.QuantumFloat
        >>> qf_1.size
        4

        """

        duplicate = copy.copy(self)

        from qrisp.core import QuantumSession

        new_qs = QuantumSession()

        # Register duplicate variable in session manager

        if name is not None:
            
            if name[-1] == "*":
                self.user_given_name = False
                name = name[:-1]
            else:
                duplicate.user_given_name = True
            
            duplicate.name = name
            new_qs.register_qv(duplicate)
            
            

        else:
            duplicate.user_given_name = False

            try:
                duplicate.name = self.name + "_dupl"
                new_qs.register_qv(duplicate)
            except NameError:
                i = 0
                while True:
                    try:
                        duplicate.name = self.name + "_dupl" + str(i)
                        new_qs.register_qv(duplicate)
                        break
                    except NameError:
                        pass
                    i += 1

        from qrisp import merge

        duplicate.qs = new_qs

        # This attribute tracks the created QuantumVariables for the
        # auto_uncompute decorator
        # We use weak references as some qrisp modules rely on reference counting
        QuantumVariable.live_qvs.append(weakref.ref(duplicate))
        duplicate.creation_time = int(self.creation_counter[0])
        duplicate.creation_counter += 1

        if qs is not None:
            merge(qs, new_qs)

        if init:
            duplicate.init_from(self)

        return duplicate

    def decoder(self, i):
        """
        The decoder method specifies how a QuantumVariable turns the outcomes of
        measurements into human-readable values. It recieves an integer ``i`` and
        returns a human-readable value.

        This method is supposed to be overloaded when defining new
        :ref:`quantum types <QuantumTypes>`.


        Parameters
        ----------
        i : int
            Integer representing the outcome of a measurement of the qubits of this
            QuantumVariable.

        Returns
        -------

            A human-readable value. Has to be hashable.

        Examples
        --------

        We create a QuantumFloat and inspect its decoder:

        >>> from qrisp import QuantumFloat
        >>> qf = QuantumFloat(3, -1, signed = False)
        >>> print(qf.decoder(1))
        0.5

        This implies that if the 3 qubits of this QuantumFloat are measured in state
        001, this outcome corresponds to the value 0.5.
        """

        from qrisp.misc import bin_rep

        return bin_rep(i, self.size)[::-1]

    def encoder(self, value):
        """
        The encoder reverses the decoder, it turns human-readable values into integers.

        If not overloaded, the encoder will perform a linear search on decoder inputs to
        match the given value.

        Parameters
        ----------
        label :
            A human-readable value.

        Raises
        ------
        Exception
            Unknown input value.

        Returns
        -------
        i : int
            The integer encoding the given value.

        Examples
        --------

        We create a QuantumChar and inspect it's encoder:

        >>> from qrisp import QuantumChar
        >>> q_ch = QuantumChar()
        >>> print(q_ch.encoder("f"))
        5

        This implies that if the 5 qubits of this QuantumChar are measured to
        ``5 = 00101``, the out come will be displayed as f.

        """
        for i in range(2**self.size):
            if self.decoder(i) == value:
                return i

        raise Exception("Value " + str(value) + " not supported by encoder.")

    def encode(self, value, permit_dirtyness = False):
        """
        The encode method allows to quickly bring a QuantumVariable in a desired
        computational basis state.

        A shorthand for this method is given by the ``[:]`` operator.

        Note that the qubits to initialize have to be fresh (i.e. no operations
        performed on them).

        Parameters
        ----------
        value :
            A value supported by the encoder.
        permit_dirtyness : bool, optional
            Surpresses the error message when calling encode on dirty qubits.

        Returns
        -------
        None.

        Examples
        --------

        We create two quantum floats and encode the value 2.5. For one of them, we
        perform an x gate onto the corresponding qubits, resulting in an error.

        >>> from qrisp import QuantumFloat, x
        >>> qf_0 = QuantumFloat(3, -1, signed = False)
        >>> qf_1 = QuantumFloat(3, -1, signed = False)
        >>> x(qf_0)
        >>> qf_0.encode(2.5)
        Exception: Tried to initialize qubits which are not fresh anymore.
        >>> qf_1[:] = 2.5
        >>> print(qf_1)
        {2.5: 1.0}

        """

        from qrisp.misc import check_if_fresh, int_encoder
        
        if not permit_dirtyness:
            if not check_if_fresh(self.reg, self.qs):
                raise Exception("Tried to initialize qubits which are not fresh anymore.")

        int_encoder(self, self.encoder(value))

    def init_state(self, state_dic):
        r"""
        The ``init_state`` method allows the initialization of arbitrary quantum states.
        It recieves a dictionary of the type

        **{value : complex number}**

        and initializes the **normalized** state. Amplitudes not specified are assumed
        to be zero.

        Note that the state initialization algorithm requires it's qubits to be in
        state $\ket{0}$.

        A shorthand for this method is the ``[:]`` operator, when handed the
        corresponding dictionary

        Parameters
        ----------
        state_dic : dict
            Dictionary describing the wave function to be initialized.

        Raises
        ------
        Exception
            Tried to initialize qubits which are not fresh anymore.

        Examples
        --------

        We create a QuantumFloat and encode the state

        .. math::

            \ket{\psi} = \sqrt{\frac{1}{3}} \ket{0.5} + i\sqrt{\frac{2}{3}} \ket{2}

        >>> from qrisp import QuantumFloat
        >>> qf = QuantumFloat(3, -1)

        We can now use either

        >>> qf.init_state({0.5: (1/3)**0.5, 2.0 : 1j*(2/3)**0.5})

        or:

        >>> qf[:] = {0.5: (1/3)**0.5, 2.0 : 1j*(2/3)**0.5}

        To acquire the expected result

        >>> print(qf)
        {2.0: 0.6667, 0.5: 0.3333}

        """

        from qrisp.misc import check_if_fresh

        if not check_if_fresh(self.reg, self.qs):
            raise Exception("Tried to initialize qubits which are not fresh anymore.")

        from qrisp import init_state

        target_array = np.zeros(2**self.size, dtype=np.complex128)

        for key in state_dic.keys():
            target_array[self.encoder(key)] = state_dic[key]

        target_array = target_array / np.vdot(target_array, target_array) ** 0.5

        init_state(self, target_array)

    def append(self, operation):
        self.qs.append(operation, self)

    def extend(self, amount, position=-1):
        """
        This method is used to add more qubits to the QuantumVariable. Using the
        position keyword it is possible to specify the position where the qubits should
        be added. By default, the qubits are added at the end.

        Parameters
        ----------
        amount : int
            The amount of qubits to add.
        position : int, optional
            The position of where to add the qubits. By default, qubits are added at the
            end.
            st of qubits which are to be added to the QuantumVariable.
            The default is None.

        Raises
        ------
        Exception
            Missmatch between proposed qubits and amount integer.

        Returns
        -------
        None.

        Examples
        --------

        We create a QuantumVariable and extend it with some extra qubits.

        >>> from qrisp import QuantumVariable
        >>> qv = QuantumVariable(3)
        >>> print(qv.reg)
        [Qubit(qv.0), Qubit(qv.1), Qubit(qv.2)]
        >>> qv.extend(3)
        >>> print(qv.reg)
        [Qubit(qv.0), Qubit(qv.1), Qubit(qv.2), Qubit(qv.6), Qubit(qv.6), Qubit(qv.6)]

        """

        if position == -1:
            position = self.size

        insertion_qubits = self.qs.request_qubits(amount)

        for i in range(amount):
            insertion_qubits[i].identifier = self.name + "." + str(self.size)
            self.reg.insert(position + i, insertion_qubits[i])
            self.size += 1

    def reduce(self, qubits, verify=False):
        r"""
        Reduces the qubit count of the QuantumVariable by removing a specified set of
        qubits.

        Parameters
        ----------
        qubits : list
            The qubits to remove from the QuantumVariable.

        verify : bool
            Boolean value which indicates wether Qrisp should verify that the reduced
            qubits are in the $\ket{0}$ state.

        Raises
        ------
        Exception
            Qubits not present in QuantumVariable.

        Exception
            Verification that the given qubits are in $\ket{0}$ state failed.

        Examples
        --------

        We create a QuantumVariable with 5 qubits and remove the first 2

        >>> from qrisp import QuantumVariable
        >>> qv = QuantumVariable(5)
        >>> print(qv.reg)
        [Qubit(qv.0), Qubit(qv.1), Qubit(qv.2), Qubit(qv.3), Qubit(qv.4)]
        >>> qv.reduce(qv[:2])
        >>> print(qv.reg)
        [Qubit(qv.2), Qubit(qv.3), Qubit(qv.4)]

        """

        try:
            len(qubits)
        except TypeError:
            qubits = [qubits]

        if not set(qubits).issubset(self.reg):
            raise Exception("Tried to reduce QuantumVariable by invalid qubits")

        # Find Qubits to be cleared
        for i in range(len(qubits)):
            for j in range(self.size):
                if self.reg[j] == qubits[i]:
                    self.reg[j].identifier = "qb_" + str(self.qs.qubit_index_counter[0])
                    self.qs.qubit_index_counter += 1
                    self.reg.pop(j)
                    break

        self.qs.clear_qubits(qubits, verify)
        # Adjust variable size
        self.size -= len(qubits)

    def get_measurement(
        self,
        plot=False,
        backend=None,
        shots=10000,
        compile=True,
        compilation_kwargs={},
        subs_dic={},
        circuit_preprocessor=None,
        filename=None,
    ):
        r"""
        Method for quick access to the measurement results of the state of the variable.
        This method returns a dictionary of the type {value : p} where p indicates the
        probability with which that value is measured.


        Parameters
        ----------
        plot : Bool, optional
            Plots the measurement results as a historgram. The default is False.
        backend : BackendClient, optional
            The backend on which to evaluate the quantum circuit. The default can be
            specified in the file default_backend.py.
        shots : integer, optional
            The amount of shots to evaluate the circuit. The default is 10000.
        compile : bool, optional
            Boolean indicating if the .compile method of the underlying QuantumSession
            should be called before. The default is True.
        compilation_kwargs  : dict, optional
            Keyword arguments for the compile method. For more details check
            :meth:`QuantumSession.compile <qrisp.QuantumSession.compile>`. The default
            is ``{}``.
        subs_dic : dict, optional
            A dictionary of Sympy symbols and floats to specify parameters in the case
            of a circuit with unspecified, :ref:`abstract parameters<QuantumCircuit>`.
            The default is {}.
        circuit_preprocessor : Python function, optional
            A function which recieves a QuantumCircuit and returns one, which is applied
            after compilation and parameter substitution. The default is None.
        filename : string, optional
            The location of where to save a generated plot. The default is None.

        Raises
        ------
        Exception
            If the containing QuantumSession is in a quantum environment, it is not
            possible to execute measurements.

        Returns
        -------
        dict
            A dictionary of values and their corresponding measurement probabilities.

        Examples
        --------

        We create an integer :ref:`QuantumFloat`, encode the value 1 and bring the qubit
        with significance 2 in superposition. We utilize the Qiskit transpiler by
        transpiling into the gate set $\{\text{CX}, \text{U}\}$

        >>> from qrisp import QuantumFloat, h
        >>> qf = QuantumFloat(3,-1)
        >>> qf[:] = 1
        >>> h(qf[2])
        >>> mes_results = qf.get_measurement(transpilation_kwargs = {"basis_gates" : ["cx", "u"]})  # noqa:501
        >>> print(mes_results)
        {1.0: 0.5, 3.0: 0.5}
        """

        if backend is None:
            if self.qs.backend is None:
                from qrisp.default_backend import def_backend

                backend = def_backend
            else:
                backend = self.qs.backend

        if len(self.qs.env_stack) != 0:
            raise Exception("Tried to get measurement within open environment")

        if self.is_deleted():
            raise Exception("Tried to get measurement from deleted QuantumVariable")

        if self.size == 0:
            return {"": 1.0}

        # Bind parameters
        if subs_dic:
            qc = self.qs.bind_parameters(subs_dic)

        else:
            if compile:
                qc = qompiler(
                    self.qs, intended_measurements=self.reg, **compilation_kwargs
                )
            else:
                qc = self.qs.copy()

        # Copy circuit in over to prevent modification
        from qrisp.quantum_network import QuantumNetworkClient

        if isinstance(backend, QuantumNetworkClient):
            self.qs.data = []
            shots = 1

        # Execute user specified circuit_preprocessor
        if circuit_preprocessor is not None:
            qc = circuit_preprocessor(qc)

        qc = qc.transpile()

        from qrisp.misc import get_measurement_from_qc

        counts = get_measurement_from_qc(qc, self.reg, backend, shots)

        # Insert outcome labels (if available and hashable)
        try:
            new_counts_dic = {}

            sorted_keys = list(counts.keys())
            sorted_keys.sort()

            for key in sorted_keys:
                new_counts_dic[self.decoder(key)] = counts[key]

            counts = new_counts_dic

            # Sort keys
            sorted_key_list = list(counts.keys())
            sorted_key_list.sort(key=lambda x: -counts[x])
            counts = {key: counts[key] for key in sorted_key_list}

        except TypeError:
            counts_tuple_list = []

            for key in counts.keys():
                counts_tuple_list.append((key, counts[key]))

            counts = counts_tuple_list

            counts.sorted(key=lambda x: x[1])

        if plot:
            outcome_labels = []
            for i in range(2**self.size):
                temp = self.decoder(i)
                
                try:
                    hash(temp)
                except TypeError:
                    raise Exception(
                        "Outcome value " + str(self.decoder(i)) + " is not hashable"
                    )

                outcome_labels.append(temp)

            plot_histogram(outcome_labels, counts, filename)
            plt.show()

        # Return dictionary of measurement results
        return counts

    def most_likely(self, **kwargs):
        """
        Performs a measurement and returns the most likely outcome.

        Parameters
        ----------
        **kwargs : Keyword arguments for the get_measurement call.

        Examples
        --------

        >>> from qrisp import QuantumFloat, ry
        >>> import numpy as np
        >>> qf = QuantumFloat(3)
        >>> ry(np.pi*9/8, qf[0])
        >>> print(qf)
        {1: 0.9619, 0: 0.0381}
        >>> qf.most_likely()
        1

        """

        return list(self.get_measurement())[0]

    def __getitem__(self, key):
        return self.reg[key]

    def __str__(self):
        return str(self.get_measurement())
    
    def __repr__(self):
        return "<" + str(type(self)).split(".")[-1][:-2] + " '" + self.name + "'>"
        return str(type(self)).split(".")[-1][:-2] + "(name = " + self.name + ")"
        return str(self)
    
    def __del__(self):
        i = 0
        while i < len(self.live_qvs):
            if self.live_qvs[i]() is None or id(self) == id(self.live_qvs[i]()):
                self.live_qvs.pop(i)
                continue
            i += 1
        

    def __len__(self):
        return self.size

    # Overload equality operator to use python syntax for if environments?
    # Not sure if the possible user confusion is worth it
    def __eq__(self, other):
        from qrisp.environments import q_eq

        return q_eq(self, other)

    def __hash__(self):
        return self.creation_time

    def __setitem__(self, key, value):
        if key != slice(None, None, None):
            raise Exception(
                "Tried to encode value into QuantumVariable using non-trivial slicing."
            )

        if isinstance(type(value), type(None)):
            return

        if isinstance(value, dict):
            self.init_state(value)
            return

        if isinstance(value, QuantumVariable):
            self.init_from(value)
            return

        self.encode(value)

    def app_phase_function(self, phi):
        r"""
        Applies a previously specified phase function to each computational basis state
        of the QuantumVariable using Gray-Synthesis.

        For a given phase function $\phi(x)$ and a QuantumVariable in state
        $\ket{\psi} = \sum_{x \in \text{Labels}} a_x \ket{x}$  this method acts as:

        .. math::

            U_{\phi} \sum_{x \in \text{Labels}} a_x \ket{x} =
            \sum_{x \in \text{Labels}} \text{exp}(i\phi(x)) a_x \ket{x}

        Parameters
        ----------
        phi : Python function
            A Python function which turns the labels of the QuantumVariable into floats.

        Examples
        --------

        We create a QuantumFloat and encode the k-th basis state of the Fourier basis.
        Finally, we will apply an inverse Fourier transformation to measure k in the
        computational basis.

        >>> import numpy as np
        >>> from qrisp import QuantumFloat, h, QFT
        >>> n = 5
        >>> qf = QuantumFloat(n, signed = False)
        >>> h(qf)

        After this, qf is in the state

        .. math::

            \ket{\text{qf}} = \frac{1}{\sqrt{2^n}} \sum_{x = 0}^{2^n} \ket{x}

        We specify phi

        >>> k = 4
        >>> def phi(x):
        >>>     return 2*np.pi*x*k/2**n

        And apply phi as a phase function

        >>> qf.app_phase_function(phi)

        qf is now in the state

        .. math::

            \ket{\text{qf}} = \frac{1}{\sqrt{2^n}} \sum_{x = 0}^{2^n}
            \text{exp}\left( \frac{2\pi ikx}{2^n}\right) \ket{x}


        Finally we apply the inverse Fourier transformation and measure:

        >>> QFT(qf, inv = True)
        >>> print(qf)
        {4: 1.0}


        """

        from qrisp.misc import app_phase_function

        app_phase_function([self], phi)

    def uncompute(self, do_it=True, recompute=False):
        """
        Method for automatic uncomputation. Uses a generalized form of
        `this algorithm <https://dl.acm.org/doi/10.1145/3453483.3454040>`_.

        For more information check the
        :ref:`uncomputation documentation<uncomputation>`.

        Parameters
        ----------
        do_it : bool, optional
            If set to False, this variable will be appended to the uncomputation stack
            of it's QuantumSession and uncomputed once an uncompute call with
            ``do_it = True`` is performed. The default is True.
        recompute : bool, optional
            If set to True, this QuantumVariable will be uncomputed but temporarily
            recomputed, if it is required for the uncomputation of another
            QuantumVariable. For more information check
            :ref:`recomputation <recomputation>`. The default is False.

        Examples
        --------

        We create two QuantumVariables, apply some gates and perform automatic
        uncomputation:

        >>> from qrisp import QuantumVariable, x, cx, h, p, mcx
        >>> a = QuantumVariable(3)
        >>> b = QuantumVariable(2)
        >>> mcx(a, b[0])
        >>> h(a[:2])
        >>> x(b[0])
        >>> cx(b[0], b[1])
        >>> p(0.5, b[1])
        >>> print(a.qs)
        QuantumCircuit:
        ---------------
                  ┌───┐
        a.0: ──■──┤ H ├───────────────
               │  ├───┤
        a.1: ──■──┤ H ├───────────────
               │  └───┘
        a.2: ──■──────────────────────
             ┌─┴─┐┌───┐
        b.0: ┤ X ├┤ X ├──■────────────
             └───┘└───┘┌─┴─┐┌────────┐
        b.1: ──────────┤ X ├┤ P(0.5) ├
                       └───┘└────────┘
        Live QuantumVariables:
        ----------------------
        QuantumVariable a
        QuantumVariable b

        >>> b.uncompute()
        >>> print(b.qs)
        QuantumCircuit:
        ---------------
             ┌────────┐                              ┌────────┐┌───┐
        a.0: ┤0       ├──────────────────────────────┤0       ├┤ H ├
             │        │                              │        │├───┤
        a.1: ┤1       ├──────────────────────────────┤1       ├┤ H ├
             │  pt3cx │                              │  pt3cx │└───┘
        a.2: ┤2       ├──────────────────────────────┤2       ├─────
             │        │┌───┐                    ┌───┐│        │
        b.0: ┤3       ├┤ X ├──■──────────────■──┤ X ├┤3       ├─────
             └────────┘└───┘┌─┴─┐┌────────┐┌─┴─┐└───┘└────────┘
        b.1: ───────────────┤ X ├┤ P(0.5) ├┤ X ├────────────────────
                            └───┘└────────┘└───┘
        Live QuantumVariables:
        ----------------------
        QuantumVariable a


        """

        if self.is_deleted():
            raise Exception("Tried to uncompute deleted QuantumVariable")

        if do_it:
            from qrisp.uncomputation import uncompute

            uncompute(self.qs, self.qs.uncomp_stack + [self], recompute)
            self.qs.uncomp_stack = []
        else:
            self.qs.uncomp_stack.append(self)

    def get_unique_name(self, name=None):
        if name is None:
            from qrisp import QuantumBool, QuantumChar, QuantumFloat

            if isinstance(self, QuantumBool):
                name = "qbl"
            elif isinstance(self, QuantumFloat):
                name = "qf"
            elif isinstance(self, QuantumChar):
                name = "qch"
            else:
                name = "qv"

        while True:
            try:
                naming_number = self.name_tracker[name]
                self.name_tracker[name] += 1
                name = name + "_" + str(naming_number)
            except KeyError:
                self.name_tracker[name] = 1
                name = name + "_0"

            i = 0
            while i < len(QuantumVariable.live_qvs):
                qv = QuantumVariable.live_qvs[i]()
                if qv is None:
                    QuantumVariable.live_qvs.pop(i)
                    continue
                if qv.name == name:
                    break
                i += 1
            else:
                break

        return name

    def init_from(self, other):
        r"""
        Method to initiate a QuantumVariable based on the state of another. This method
        does NOT copy the state. Much rather it performs the operation


        .. math::

            U_{\text{init_from}} \left( \sum_{x \in \text{labels}} a_x \ket{x}
            \right)  \ket{0} = \sum_{x \in \text{labels}} a_x \ket{x} \ket{x}


        This is different from a state copying operation:

        .. math::

            U_{\text{copy}} \left( \sum_{x \in \text{labels}} a_x \ket{x} \right)
            \ket{0} = \left( \sum_{x \in \text{labels}} a_x \ket{x} \right)
            \left( \sum_{x \in \text{labels}} a_x \ket{x} \right)


        A shorthand for initiating this way is the ``[:]`` operator.


        Parameters
        ----------
        other : QuantumVariable
            The QuantumVariable from which to initiate.

        Raises
        ------
        Exception
            Tried to initialize qubits which are not fresh anymore.

        Examples
        --------

        We create a QuantumFloat, and bring it into superposition.

        >>> from qrisp import QuantumFloat, h, multi_measurement
        >>> qf_a = QuantumFloat(8)
        >>> qf_a[:] = 6
        >>> h(qf_a[0])
        >>> print(qf_a)
        {6: 0.5, 7: 0.5}

        We now duplicate and initiate the duplicate

        >>> qf_b = qf_a.duplicate()
        >>> print(qf_b)
        {0: 1.0}
        >>> qf_b.init_from(qf_a)
        >>> print(multi_measurement([qf_a, qf_b]))
        {(6, 6): 0.5, (7, 7): 0.5}

        The slicing operator achieves the same:

        >>> qf_c = qf_a.duplicate()
        >>> qf_c[:] = qf_b
        >>> print(multi_measurement([qf_a, qf_b, qf_c]))
        {(6, 6, 6): 0.5, (7, 7, 7): 0.5}



        """

        if not type(self) == type(other):
            raise Exception(
                "Tried to initialize " + str(type(self)) + " from " + str(type(other))
            )

        from qrisp.misc import check_if_fresh

        if not check_if_fresh(self.reg, self.qs):
            raise Exception("Tried to initialize qubits which are not fresh anymore")

        self.qs.cx(other.reg, self.reg)

    @classmethod
    def custom(self, label_list, decoder=None, qs=None, name=None):
        """
        Creates a QuantumVariable with customized outcome labels.

        Note that this is a class method, implying there is no need to create another
        QuantumVariable first to call this method.

        Parameters
        ----------
        label_list : list
            A list of outcome labels.
        decoder : function, optional
            The decoder function. If given none, the labels will be encoded according to
            their placement in the ``label_list``.
        qs : QuantumSession, optional
            The :ref:`QuantumSession` in which to register the customized
            QuantumVariable. If given none, the QuantumVariable will be registered in a
            new QuantumSession.
        name : string, optional
            The name of the QuantumVariable. If given none, a suited name will be
            generated.

        Returns
        -------
        CustomQuantumVariable
            A QuantumVariable with the desired outcome labels.

        Examples
        --------

        We create a QuantumVariable with some examples values as outcome labels and
        bring it into uniform superposition.

        >>> from qrisp import QuantumVariable, h
        >>> qv = QuantumVariable.custom(["lorem", "ipsum", "dolor", "sit", 42, (1,2,3)])
        >>> h(qv)
        >>> print(qv)
        {'lorem': 0.125, 'ipsum': 0.125, 'dolor': 0.125, 'sit': 0.125, 42: 0.125,
        (1, 2, 3): 0.125, 'undefined_label_6': 0.125, 'undefined_label_7': 0.125}

        """
        from qrisp.misc import custom_qv

        return custom_qv(label_list, decoder=decoder, qs=qs, name=name)


def plot_histogram(outcome_labels, counts, filename=None):
    res_list = []

    for k in range(len(outcome_labels)):
        try:
            res_list.append(counts[outcome_labels[k]])
        except KeyError:
            res_list.append(0)

    plt.bar(outcome_labels, res_list, width = 0.8/len(outcome_labels))
    plt.grid()
    plt.ylabel("Measurement probability")
    plt.xlabel("QuantumVariable value")

    if filename:
        plt.savefig(filename, dpi=400, bbox_inches="tight")
    else:
        plt.show()



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

# Created by ann81984 at 05.05.2022
import numpy as np

from qrisp import QuantumFloat, QuantumChar, QFT, multi_measurement, as_hamiltonian, h


def test_diagonal_hamiltonian_application():
    # Hamiltonian function
    # This function should take an element from the outcome labels of qf and turns it into a phase
    # In this case we are handling a QuantumFloat, so the input is a float
    # Another case could be where we are handling a QuantumChar
    # In this case the function should process a char
    def hamiltonian(x):
        return np.pi * np.sin(x**2 * np.pi * 2) * x
        # If we were processing a QuantumChar, the hamiltonian could look like this:
        # if x == "a":
        #     return 0.5*np.pi
        # elif x == "b":
        #     return 0.25*np.pi
        # else:
        #     return np.pi

    # Create quantum session
    # qs = QuantumSession()

    # Create quantum float
    qf = QuantumFloat(5, -5)

    # Bring qf in uniform superposition state in order to be able to see which phase has been applied to which state
    h(qf)

    # Apply hamiltonian
    qf.app_phase_function(hamiltonian)

    # Simulate phases
    sv_ = qf.qs.statevector("function")
    sv = lambda x: sv_({qf: x})

    def sv_phase(label):
        return np.angle(sv(label)) - np.angle(sv(x[0])) + hamiltonian(x[0])

    # Prepare x array (for plotting)
    x = np.array([qf.decoder(i) for i in range(2**qf.size)])
    sv_phase_array = np.array([sv_phase(i) for i in x])

    # Correct global phase
    # sv_phases += (-sv_phases[0] + hamiltonian(x[0]))

    # Check if phases agree
    from numpy.linalg import norm

    # assert abs((sv_phase(i) - hamiltonian(i))%(2*np.pi)) < 1E-4
    assert norm((sv_phase_array - hamiltonian(x))) < 1e-4
    # Plot results
    # import matplotlib.pyplot as plt
    # plt.plot(x, hamiltonian(x) % (2 * np.pi), "o", label="Hamiltonian")
    # plt.plot(x, sv_phase_array % (2 * np.pi), ".", label="Simulated phases")
    #
    # plt.ylabel("Phase [radian]")
    # plt.xlabel("Quantum Float outcome labels")
    # plt.grid()
    # plt.legend()
    # plt.show()

    ###################
    # In this example we will demonstrate how a phase function with multiple arguments can be synthesized
    # For this we will create a phase function which encodes the fourier transform of different integers on the QuantumFloat x
    # conditioned on the value of a QuantumChar c
    # We will then apply the inverse Fourier transform to x and measure the results

    x_size = 3

    # Create Variables
    x = QuantumFloat(x_size, 0, signed=False)
    c = QuantumChar()

    # Bring x into uniform superposition so the phase function application yields a fourier transformed computation basis state
    h(x)

    # Bring c into partial superposition (here |a> + |b> + |c> + |d>)
    h(c[0])
    h(c[1])

    # In order to define the hamiltonian, we can use regular Python syntax
    # The decorator "as_hamiltonian" turns it into a function that takes Quantum Variables as arguments
    # The decorator will add the keyword argument t to the function which mimics the t in exp(i*H*t)
    @as_hamiltonian
    def apply_multi_var_hamiltonian(c, x):
        if c == "a":
            k = 2
        elif c == "b":
            k = 2
        elif c == "c":
            k = 3
        else:
            k = 4

        # Return phase value
        # This is the phase distribution of the Fourier-transform
        # of the computational basis state |k>
        return k * x * 2 * np.pi / 2**x_size

    # Apply Hamiltonian
    apply_multi_var_hamiltonian(c, x, t=1)

    # Apply inverse Fourier transform
    QFT(x, inv=True)

    # Acquire measurement results
    assert multi_measurement([c, x]) == {
        ("a", 2): 0.25,
        ("b", 2): 0.25,
        ("c", 3): 0.25,
        ("d", 4): 0.25,
    }



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

# -*- coding: utf-8 -*-

# This file contains some functions that enhance the speed
# of SparseBiArray manipulation

import numpy as np
from numba import int64, njit, prange, vectorize


# It can happen that the coo-matrix multiplication puts two data entries
# for the same index. These function recognizes and sums these duplicates
def sum_duplicates(nz_indices, data):
    new_indices, reconstr_indices = np.unique(nz_indices, return_inverse=True)

    if len(nz_indices) > 1e4:
        new_data = sum_duplicates_aux_jitted(new_indices, data, reconstr_indices)
    else:
        new_data = sum_duplicates_aux(new_indices, data, reconstr_indices)

    return new_indices, new_data


def sum_duplicates_aux(new_indices, old_data, reconstr_indices):
    new_data = np.zeros(new_indices.shape, dtype=old_data.dtype)

    for i in range(len(reconstr_indices)):
        new_data[reconstr_indices[i]] += old_data[i]

    return new_data


@njit(nogil=True, cache=True)
def sum_duplicates_aux_jitted(new_indices, old_data, reconstr_indices):
    new_data = np.zeros(new_indices.shape, dtype=old_data.dtype)

    for i in range(len(reconstr_indices)):
        new_data[reconstr_indices[i]] += old_data[i]

    return new_data


# This function rounds entries, which are below a certain float thresh and removes the
# tracked indices.
# @njit(parallel = True)
def elim_zeros(nz_indices, data, decimals=8):
    # return nz_indices, data
    data = np.around(data, decimals=decimals, out=data)

    # mask = np.abs(data) > 10**-decimals
    mask = np.nonzero(data)[0]

    return nz_indices[mask], data[mask]


# This function applies an index permutation onto
# flattened indices.
# Consider the situation that we have the two flattened indices
# [1,3]
# of a 2x2 matrix.
# If we now want to permute the axes, we permute the bit level
# representation of these numbers
# ["01", "10"] ==> ["10", "01"]


@njit(nogil=True, cache=True)
def permute_axes(index, permutation):
    permutation = -permutation + len(permutation) - 1
    permutation = permutation[::-1]
    new_index = np.zeros(index.size, dtype=np.int64)
    for i in range(len(permutation)):
        if permutation[i] == i:
            # new_index = new_index | (index&(1<<permutation[i]))
            for j in prange(len(new_index)):
                new_index[j] = new_index[j] | (index[j] & (1 << permutation[i]))
        else:
            # new_index = new_index |
            # (((index&(1<<permutation[i]))>>permutation[i])<<(i))
            for j in prange(len(new_index)):
                new_index[j] = new_index[j] | (
                    ((index[j] & (1 << permutation[i])) >> permutation[i]) << (i)
                )

    return new_index


# This function retrieves the coordinates from
# flattened coordinades for a given shape of length two
def get_coordinates(indices, shape):
    return get_coordinates_0(indices, int(np.log2(shape[1]))), get_coordinates_1(
        indices, int(np.log2(shape[1]))
    )


@vectorize([int64(int64, int64)])
def get_coordinates_0(index, bit_shape_1):
    return index >> bit_shape_1


@vectorize([int64(int64, int64)])
def get_coordinates_1(index, bit_shape_0):
    return index & ((1 << bit_shape_0) - 1)


# This function basically inverts get_coordinates.
# Ie. it generates flat coordinates from row and column indices
def gen_flat_coords(col, row, shape):
    return gen_flat_coords_vec(col, row, int(np.log2(shape[1])))


@vectorize([int64(int64, int64, int64)], cache=True)
def gen_flat_coords_vec(col, row, bit_shape_1):
    return row << bit_shape_1 | col


# This function constructs a flattened numpy array from
# given flattened coordinates and data
@njit(nogil=True, cache=True)
def construct_flat_array(flat_coords, data, size):
    res = np.zeros(size, dtype=data.dtype)

    for i in range(len(flat_coords)):
        res[flat_coords[i]] += data[i]

    return res


@njit(cache=True)
def extract_bit_range(index, start, stop):
    breadth = stop - start
    return (index & (((1 << breadth) - 1) << start)) >> start


@njit(cache=True)
def insert_bit_range(target, bit_string, start):
    target |= bit_string << start


@njit(nogil=True, cache=True)
def jitted_permuter(index, new_index, bit_partition, permuted_bit_partition, perm):
    # for i in range(len(perm)):
    #     # bit_range = extract_bit_range(index, bit_partition[i], bit_partition[i+1])
    #     # insert_bit_range(new_index, bit_range, permuted_bit_partition[perm[i]])
    #     #Inlined version more efficient:
    #     new_index |= (((index&((((1<<(bit_partition[i+1] -
    #     bit_partition[i]))-1))<<bit_partition[i]))>>bit_partition[i])<<
    #     permuted_bit_partition[perm[i]])

    for j in prange(index.size):
        for i in range(len(perm)):
            # bit_range = extract_bit_range(index, bit_partition[i], bit_partition[i+1])
            # insert_bit_range(new_index, bit_range, permuted_bit_partition[perm[i]])
            # Inlined version more efficient:
            new_index[j] |= (
                (
                    index[j]
                    & (
                        ((1 << (bit_partition[i + 1] - bit_partition[i])) - 1)
                        << bit_partition[i]
                    )
                )
                >> bit_partition[i]
            ) << permuted_bit_partition[perm[i]]


def permuter(index, new_index, bit_partition, permuted_bit_partition, perm):
    for i in range(len(perm)):
        # bit_range = extract_bit_range(index, bit_partition[i], bit_partition[i+1])
        # insert_bit_range(new_index, bit_range, permuted_bit_partition[perm[i]])
        # Inlined version more efficient:
        new_index |= (
            (
                index
                & (
                    ((1 << (bit_partition[i + 1] - bit_partition[i])) - 1)
                    << bit_partition[i]
                )
            )
            >> bit_partition[i]
        ) << permuted_bit_partition[perm[i]]


def invert_perm(perm):
    return [perm.index(i) for i in range(len(perm))]


def permute_axes(index, index_bit_permutation):
    n = len(index_bit_permutation)

    index_bit_permutation = -np.array(index_bit_permutation) + n - 1
    index_bit_permutation = list(index_bit_permutation)[::-1]

    index_bit_permutation = list(index_bit_permutation)

    # Get the permutation of the axes

    perm = [index_bit_permutation.index(i) for i in range(n)]

    # Initialize the shape
    perm_shape = n * [1]

    i = 1
    while i < len(perm):
        # If this condition is met, both indices form an interval,
        # that does not need further permutation
        if perm[i] - 1 == perm[i - 1]:
            # Remove the index from the perm and perm_shape list
            perm.pop(i)
            perm_shape.pop(i)

            # Update the shape
            perm_shape[i - 1] += 1
            # Update the permutation indices
            for j in range(len(perm)):
                if perm[j] > perm[i - 1]:
                    perm[j] -= 1
        else:
            i += 1

    # print(perm)
    new_index = np.zeros(index.size, dtype=np.int64)

    bit_partition = [sum(perm_shape[:i]) for i in range(len(perm_shape) + 1)]

    # print(bit_partition)
    perm_shape_permuted = [perm_shape[i] for i in invert_perm(perm)]
    permuted_bit_partition = [
        sum(perm_shape_permuted[:i]) for i in range(len(perm_shape) + 1)
    ]

    if len(perm) * index.size > 2**10:
        jitted_permuter(
            index,
            new_index,
            np.array(bit_partition),
            np.array(permuted_bit_partition),
            np.array(perm),
        )
    else:
        permuter(
            index,
            new_index,
            np.array(bit_partition),
            np.array(permuted_bit_partition),
            np.array(perm),
        )

    return new_index


@njit(cache=True)
def invert_permutation(p):
    """Return an array s with which np.array_equal(arr[p][s], arr) is True.
    The array_like argument p must be some permutation of 0, 1, ..., len(p)-1.
    """
    s = np.empty(len(p), dtype=np.int64)
    s[p] = np.arange(p.size)
    return s


@njit(nogil=True, cache=True)
def bi_array_moveaxis(data_array, index_perm, f_index_array):
    for i in range(len(index_perm)):
        shifted_value = 1 << index_perm[i]
        start_index = 1 << i

        for j in range(start_index):
            f_index_array[start_index ^ j] = shifted_value ^ f_index_array[j]

    # return f_index_array
    return data_array[f_index_array]



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


from qrisp.circuit import QuantumCircuit
from qrisp.environments import QuantumEnvironment


class GateWrapEnvironment(QuantumEnvironment):
    """
    This environment allows to hide complexity in the circuit visualisation.
    Operations appended inside this environment are bundled into a single
    :ref:`Instruction` object.

    The functionality of this :ref:`QuantumEnvironment` can also be used with the
    :meth:`gate_wrap <qrisp.gate_wrap>` decorator.

    After compiling, the wrapped instruction can be retrieved using the
    ``.instruction`` attribute.

    Parameters
    ----------
    name : string, optional
        The name of the resulting gate. The default is None.

    Examples
    --------

    We create some :ref:`QuantumVariable` and execute some gates inside
    a GateWrapEnvironment: ::

        from qrisp import QuantumVariable, GateWrapEnvironment, x, y, z

        qv = QuantumVariable(3)

        gwe = GateWrapEnvironment("example")


        with gwe:
            x(qv[0])
            y(qv[1])

        z(qv[2])


    >>> print(qv.qs)
    QuantumCircuit:
    ---------------
          ┌──────────┐
    qv.0: ┤0         ├
          │  example │
    qv.1: ┤1         ├
          └──┬───┬───┘
    qv.2: ───┤ Z ├────
             └───┘
    Live QuantumVariables:
    ----------------------
    QuantumVariable qv

    We can access the instruction, which has been appended using the
    ``.instruction`` attribute:

    >>> instruction = gwe.instruction
    >>> print(instruction.op.definition)
           ┌───┐
    qb_41: ┤ X ├
           ├───┤
    qb_42: ┤ Y ├
           └───┘

    Using the :meth:`gate_wrap <qrisp.gate_wrap>` decorator we can quickly gate wrap
    functions: ::

        from qrisp import gate_wrap

        @gate_wrap
        def example_function(qv):
            x(qv[0])
            y(qv[1])
            z(qv[2])


        example_function(qv)

    >>> print(qv.qs)
    QuantumCircuit:
    ---------------
          ┌──────────┐┌───────────────────┐
    qv.0: ┤0         ├┤0                  ├
          │  example ││                   │
    qv.1: ┤1         ├┤1 example_function ├
          └──┬───┬───┘│                   │
    qv.2: ───┤ Z ├────┤2                  ├
             └───┘    └───────────────────┘
    Live QuantumVariables:
    ----------------------
    QuantumVariable qv

    """

    def __init__(self, name=None):
        super().__init__()
        self.name = name

        self.manual_allocation_management = True

    def compile(self):
        temp_data_list = list(self.env_qs.data)

        self.env_qs.data = []
        super().compile()

        compiled_qc = self.env_qs.clearcopy()

        compiled_qc.data = list(self.env_qs.data)

        self.env_qs.clear_data()
        self.env_qs.data.extend(temp_data_list)

        if len(compiled_qc.data) == 0:
            self.instruction = None
            return None

        qc = QuantumCircuit(len(self.env_qs.qubits), len(self.env_qs.clbits))

        translation_dic = {
            self.env_qs.qubits[i]: qc.qubits[i] for i in range(len(qc.qubits))
        }
        translation_dic.update(
            {self.env_qs.clbits[i]: qc.clbits[i] for i in range(len(qc.clbits))}
        )

        qubit_set = set([])

        dealloc_list = []
        alloc_list = []

        for instr in compiled_qc.data:
            qubit_set = qubit_set.union(
                set([translation_dic[qb] for qb in instr.qubits])
            )
            if instr.op.name == "qb_dealloc":
                instr.qubits[0].allocated = True
                dealloc_list.append(instr)
                continue
            if instr.op.name == "qb_alloc":
                alloc_list.append(instr)

            qc.append(
                instr.op,
                [translation_dic[qb] for qb in instr.qubits],
                [translation_dic[cb] for cb in instr.clbits],
            )

        idle_qubit_list = list(set(qc.qubits) - qubit_set)

        for j in range(len(idle_qubit_list)):
            for i in range(len(qc.qubits)):
                if qc.qubits[i].identifier == idle_qubit_list[j].identifier:
                    qc.qubits.pop(i)
                    break

        translation_dic_inv = {
            translation_dic[key]: key for key in translation_dic.keys()
        }

        gate = qc.to_gate(self.name)

        for instr in alloc_list:
            self.env_qs.append(instr)

        self.env_qs.append(
            gate,
            [translation_dic_inv[qb] for qb in qc.qubits],
            [translation_dic_inv[cb] for cb in qc.clbits],
        )
        self.instruction = self.env_qs.data[-1]

        for instr in dealloc_list:
            self.env_qs.append(instr)
            instr.qubits[0].allocated = False



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


import numpy as np
from qrisp import QuantumArray, QuantumVariable, gate_wrap, gphase, h, mcp, mcz, x, merge, recursive_qs_search, IterationEnvironment


# Applies the grover diffuser onto the (list of) quantum variable input_object
def diffuser(input_object, phase=np.pi):
    """
    Applies the Grover diffuser onto (multiple) QuantumVariables.

    Parameters
    ----------
    input_object : QuantumVariable or list[QuantumVariable]
        The (list of) QuantumVariables to apply the Grover diffuser on.

    Examples
    --------

    We apply the Grover diffuser onto several QuantumChars:

    >>> from qrisp import QuantumChar
    >>> from qrisp.grover import diffuser
    >>> q_ch_list = [QuantumChar(), QuantumChar(), QuantumChar()]
    >>> diffuser(q_ch_list)
    >>> print(q_ch_list[0].qs)
              ┌────────────┐
    q_ch_0.0: ┤0           ├
              │            │
    q_ch_0.1: ┤1           ├
              │            │
    q_ch_0.2: ┤2           ├
              │            │
    q_ch_0.3: ┤3           ├
              │            │
    q_ch_0.4: ┤4           ├
              │            │
    q_ch_1.0: ┤5           ├
              │            │
    q_ch_1.1: ┤6           ├
              │            │
    q_ch_1.2: ┤7  diffuser ├
              │            │
    q_ch_1.3: ┤8           ├
              │            │
    q_ch_1.4: ┤9           ├
              │            │
    q_ch_2.0: ┤10          ├
              │            │
    q_ch_2.1: ┤11          ├
              │            │
    q_ch_2.2: ┤12          ├
              │            │
    q_ch_2.3: ┤13          ├
              │            │
    q_ch_2.4: ┤14          ├
              └────────────┘
    """

    if isinstance(input_object, QuantumArray):
        input_object = [qv for qv in input_object.flatten()]

    if isinstance(input_object, list):
        [h(qv) for qv in input_object]
        tag_state(
            {qv: "0" * qv.size for qv in input_object}, binary_values=True, phase=phase
        )
        [h(qv) for qv in input_object]
        gphase(np.pi, input_object[0][0])
    else:
        h(input_object)
        tag_state(
            {input_object: input_object.size * "0"}, binary_values=True, phase=phase
        )
        h(input_object)
        gphase(np.pi, input_object[0])

    return input_object


def tag_state(tag_specificator, binary_values=False, phase=np.pi):
    """
    Applies a phase tag to (multiple) QuantumVariables. The tagged state is specified in
    the dictionary ``tag_specificator``. This dictionary should contain the
    QuantumVariables as keys and the labels of the states which should be tagged as
    values.


    Parameters
    ----------
    tag_specificator : dict
        A dictionary specifying which state should be tagged.
    binary_values : bool, optional
        If set to True, the values in the tag_specificator dict have to be bitstrings
        instead of labels. The default is False.
    phase : float or sympy.Symbol, optional
        Specify the phase shift the tag should perform. The default is pi, i.e. a
        multi-controlled Z gate.


    Examples
    --------

    We construct an oracle that tags the states -3 and 2 on two QuantumFloats ::

        from qrisp.grover import tag_state

        def test_oracle(qf_list):

            tag_dic = {qf_list[0] : -3, qf_list[1] : 2}

            tag_state(tag_dic)

    """

    qv_list = list(tag_specificator.keys())

    states = [tag_specificator[qv] for qv in qv_list]

    if not len(states):
        states = ["1" * qv.size for qv in qv_list]

    # The phase tag will be executed by a multicontrolled z-gate
    # Even though the multicontrolled z-gate has no prefered qubit,
    # in qiskit we have to give one "base" qubit and the other qubits are
    # the control qubits. For the control qubits we can easily specify the required
    # control state, however for the base qubit it's not that easy.
    # If our base qubit should be in state 0 we therefore have to apply an x gate
    # on it, execute the controlled z gate and then apply another x gate

    # Apply x gate (if neccessary)

    bit_string = ""
    from qrisp.misc import bin_rep

    for i in range(len(qv_list)):
        if binary_values:
            bit_string += states[i][::-1]
        else:
            bit_string += bin_rep(qv_list[i].encoder(states[i]), qv_list[i].size)[::-1]

    qubit_list = sum([list(qv.reg) for qv in qv_list], [])
    state = bit_string

    if state[-1] == "0":
        x(qubit_list[-1])

    # The control state is the state we are looking for without the base qubit
    ctrl_state = state[:-1]
    if phase == np.pi:
        mcz(qubit_list, ctrl_state=ctrl_state + "1")
    else:
        mcp(phase, qubit_list, ctrl_state=ctrl_state + "1")

    # Apply the final x gate
    if state[-1] == "0":
        x(qubit_list[-1])


def grovers_alg(
    qv_list,
    oracle_function,
    kwargs={},
    iterations=0,
    winner_state_amount=None,
    exact=False,
):
    r"""
    Applies Grover's algorithm to a given oracle (in the form of a Python function).

    Parameters
    ----------
    qv_list : QuantumVariable or list[QuantumVariable]
        A (list of) QuantumVariables on which to execute Grover's algorithm.
    oracle_function : function
        A Python function tagging the winner states.
    kwargs : dict, optional
        A dictionary containing keyword arguments for the oracle. The default is {}.
    iterations : int, optional
        The amount of Grover iterations to perfrom.
    winner_state_amount : int, optional
        If not given the amount of iterations, the established formula will be used
        based on the amount of winner states. The default is 1.
    exact : bool, optional
        If set to True, the `exact version <https://arxiv.org/pdf/quant-ph/0106071.pdf>`
        of Grover's algorithm will be evaluated. For this, the correct
        ``winner_state_amount`` has to be supplied and the oracle has to support the
        keyword argument ``phase`` which specifies how much the winner states are
        phaseshifted (in standard Grover this would be $\pi$).
        
    Raises
    ------
    
    Exception
        Applied oracle introducing new QuantumVariables without uncomputing/deleting

    Examples
    --------

    We construct an oracle that tags the states -3 and 2 on two QuantumFloats and apply
    Grover's algorithm to it. ::

        from qrisp import QuantumFloat

        #Create list of QuantumFloats
        qf_list = [QuantumFloat(2, signed = True), QuantumFloat(2, signed = True)]

        from qrisp.grover import tag_state, grovers_alg

        def test_oracle(qf_list):

            tag_dic = {qf_list[0] : -3, qf_list[1] : 2}
            tag_state(tag_dic)

        grovers_alg(qf_list, test_oracle)

    >>> from qrisp.misc import multi_measurement
    >>> print(multi_measurement(qf_list))
    {(-3, 2): 0.9966, (0, 0): 0.0001, (0, 1): 0.0001, (0, 2): 0.0001, (0, 3): 0.0001,
    (0, -4): 0.0001, (0, -3): 0.0001, (0, -2): 0.0001, (0, -1): 0.0001, (1, 0): 0.0001,
    (1, 1): 0.0001, (1, 2): 0.0001, (1, 3): 0.0001, (1, -4): 0.0001, (1, -3): 0.0001,
    (1, -2): 0.0001, (1, -1): 0.0001, (2, 0): 0.0001, (2, 1): 0.0001, (2, 2): 0.0001,
    (2, 3): 0.0001, (2, -4): 0.0001, (2, -3): 0.0001, (2, -2): 0.0001, (2, -1): 0.0001,
    (3, 0): 0.0001, (3, 1): 0.0001, (3, 2): 0.0001, (3, 3): 0.0001, (3, -4): 0.0001,
    (3, -3): 0.0001, (3, -2): 0.0001, (3, -1): 0.0001, (-4, 0): 0.0001, (-4, 1): 0.0001,
     (-4, 2): 0.0001, (-4, 3): 0.0001, (-4, -4): 0.0001, (-4, -3): 0.0001,
     (-4, -2): 0.0001, (-4, -1): 0.0001, (-3, 0): 0.0001, (-3, 1): 0.0001,
     (-3, 3): 0.0001, (-3, -4): 0.0001, (-3, -3): 0.0001, (-3, -2): 0.0001,
     (-3, -1): 0.0001, (-2, 0): 0.0001, (-2, 1): 0.0001, (-2, 2): 0.0001,
     (-2, 3): 0.0001, (-2, -4): 0.0001, (-2, -3): 0.0001, (-2, -2): 0.0001,
     (-2, -1): 0.0001, (-1, 0): 0.0001, (-1, 1): 0.0001, (-1, 2): 0.0001,
     (-1, 3): 0.0001, (-1, -4): 0.0001, (-1, -3): 0.0001, (-1, -2): 0.0001,
     (-1, -1): 0.0001}

    **Exact Grovers Algorithm**

    In the next example, we will showcase the ``exact`` functionality. For this we
    create an oracle, which tags all the states of a QuantumVariable, that contain 3
    ones and 2 zeros.

    To count the amount of ones we use quantum phase estimation on the operator

    .. math::

        U = \text{exp}\left(\frac{i 2 \pi}{2^k}
        \sum_{i = 0}^{n-1} ( 1 - \sigma_{z}^i )\right)


    ::

        from qrisp import QPE, p, QuantumVariable, lifted
        from qrisp.grover import grovers_alg, tag_state
        import numpy as np

        def U(qv, prec = None, iter = 1):
            for i in range(qv.size):
                p(iter*2*np.pi/2**prec, qv[i])

        @lifted
        def count_ones(qv):
            prec = int(np.ceil(np.log2(qv.size+1)))
            res = QPE(qv, U, precision = prec, iter_spec = True, kwargs = {"prec" : prec})
            return res<<prec


    Quick test:

    >>> qv = QuantumVariable(5)
    >>> qv[:] = {"11000" : 1, "11010" : 1, "11110" : 1}
    >>> count_qf = count_ones(qv)
    >>> count_qf.qs.statevector()
    sqrt(3)*(|11000>*|2> + |11010>*|3> + |11110>*|4>)/3

    We now define the oracle ::

        def counting_oracle(qv, phase = np.pi, k = 1):

            count_qf = count_ones(qv)

            tag_state({count_qf : k}, phase = phase)

            count_qf.uncompute()

    And evaluate Grover's algorithm ::

        n = 5
        k = 3
        qv = QuantumVariable(n)

        import math

        grovers_alg(qv, counting_oracle, exact = True, winner_state_amount = math.comb(n,k), kwargs = {"k" : k})  # noqa


    >>> print(qv)
    {'11100': 0.1,
     '11010': 0.1,
     '10110': 0.1,
     '01110': 0.1,
     '11001': 0.1,
     '10101': 0.1,
     '01101': 0.1,
     '10011': 0.1,
     '01011': 0.1,
     '00111': 0.1}

    We see that contrary to regular Grover's algorithm, the states which have not been
    tagged by the oracle have 0 percent measurement probability.

    """

    if exact and winner_state_amount is None:
        raise Exception(
            "Tried to call exact Grover's algorithm without specifying "
            "winner_state_amount"
        )
    elif winner_state_amount is None:
        winner_state_amount = 1

    if isinstance(qv_list, list):
        N = 2 ** sum([qv.size for qv in qv_list])
    elif isinstance(qv_list, QuantumArray):
        N = 2 ** sum([qv.size for qv in qv_list.flatten()])
    elif isinstance(qv_list, QuantumVariable):
        N = 2**qv_list.size

    if exact:
        # Implementation for phase calculation for exact grovers alg as in
        # https://arxiv.org/pdf/quant-ph/0106071.pdf
        iterations = 1
        tmp = (
            np.sin(np.pi / (4 * (iterations - 1) + 6))
            * (N / winner_state_amount) ** 0.5
        )

        while tmp > 1:
            iterations += 1
            tmp = (
                np.sin(np.pi / (4 * (iterations - 1) + 6))
                * (N / winner_state_amount) ** 0.5
            )

        phi = 2 * np.arcsin(
            np.sin(np.pi / (4 * (iterations - 1) + 6))
            * (N / winner_state_amount) ** 0.5
        )

    else:
        if iterations == 0:
            iterations = np.pi / 4 * np.sqrt(N / winner_state_amount)
            iterations = int(np.round(iterations))

    for qv in qv_list:
        h(qv)

    merge(qv_list)
    qs = recursive_qs_search(qv_list)[0]
    
    qv_amount = len(qs.qv_list)

    with IterationEnvironment(qs, iterations):
    # for i in range(iterations):
        if exact:
            oracle_function(qv_list, phase=phi, **kwargs)
            diffuser(qv_list, phase=phi)
        else:
            oracle_function(qv_list, **kwargs)
            diffuser(qv_list)
            
        if qv_amount != len(qs.qv_list):
            raise Exception("Applied oracle introducing new QuantumVariables without uncomputing/deleting")


# Workaround to keep the docstring but still gatewrap

temp = diffuser.__doc__

diffuser = gate_wrap(permeability=[], is_qfree=False)(diffuser)

diffuser.__doc__ = temp


temp = tag_state.__doc__

tag_state = gate_wrap(permeability="args", is_qfree=True)(tag_state)

tag_state.__doc__ = temp
